from tempfile import NamedTemporaryFile
from lacore.crypt import CryptIO
from lacore.cipher import get_cipher
from shutil import copyfileobj
from zipfile import ZipFile
from abc import ABCMeta, abstractmethod


def restore_archive(archive, path, cert, folder, tmpdir, cb=None):
    cipher = get_cipher(archive, cert)
    with open(path, 'rb') as infile:
        with NamedTemporaryFile() as dst:
            with CryptIO(infile, cipher) as cf:
                copyfileobj(cf, dst)
            dst.flush()
            dst.seek(0)
            with ZipFile(dst) as zf:
                map(cb,
                    map(lambda zi: zf.extract(zi, folder), zf.infolist()))


class Archiver(object):
    __metaclass__ = ABCMeta

    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def args(self, items, cb=None):
        """ process each item before archiving, calling
            callback if supplied
        """

    @abstractmethod
    def archive(self, items, dest, cb=None):
        """ write every item to test and yield the item """



import io

from .zip import ZipArchiver
from functools import partial
from collections import OrderedDict


class ProgressStream(io.RawIOBase):
    def __init__(self, fp, progress=None):
        self.fp = fp
        self.progress = progress
        self.leftover = None

    def readable(self):
        return True

    def readinto(self, b):
        l = len(b)  # We're supposed to return at most this much
        chunk = self.leftover or self.fp.read(l)
        if chunk is None:
            return None
        output, self.leftover = chunk[:l], chunk[l:]
        b[:len(output)] = output
        if self.progress is not None:
            self.progress(output)
        return len(output)

    def close(self):
        super(ProgressStream, self).close()
        self.fp.close()


class StreamArchiver(ZipArchiver):
    processed = None
    stream_chunk = io.DEFAULT_BUFFER_SIZE

    def process(self, rel, data):
        self.processed[rel] += len(data)

    def __init__(self, stream_chunk=None, progress_cb=None, *args, **kwargs):
        if stream_chunk is not None:
            self.stream_chunk = stream_chunk
        self.progress_cb = progress_cb
        super(StreamArchiver, self).__init__(*args, **kwargs)

    def args(self, items, cb):
        self.processed = OrderedDict()
        for stream, rel in items:
            self.processed[rel] = 0
            if cb is not None:
                cb(stream, rel)
            s = io.BufferedReader(
                ProgressStream(
                    stream, partial(self.process, rel)),
                self.stream_chunk)
            yield ((s,), {'arcname': rel.encode('utf-8')})



import yaml
import json
import re
import mmap

try:
    import pyaml
except ImportError:
    pyaml = None

from lacore.exceptions import InvalidArchiveError
from lacore.date import parse_timestamp, format_timestamp
from datetime import datetime
from base64 import b64encode, b64decode
from .elements import (Archive, Meta, Certificate, Auth,
                       Cipher, Links, Signature)
from .serialize import load_file


def make_adf(archive=None, out=None, pretty=False):
    """
    >>> import sys
    >>> meta = Meta('zip', 'aes-256-ctr', created='now')
    >>> archive = Archive('title', meta)
    >>> cert = Certificate(chr(255)*16)
    >>> make_adf([archive, cert], out=sys.stdout, pretty=True)
    !archive
    meta: !meta
      cipher: aes-256-ctr
      created: now
      format: zip
    title: title
    --- !certificate
    key: !!binary |
      /////////////////////w==
    >>> make_adf([archive, cert], out=sys.stdout, pretty=True)
    !archive
    meta: !meta
      cipher: aes-256-ctr
      created: now
      format: zip
    title: title
    --- !certificate
    key: !!binary |
      /////////////////////w==
    """

    if not hasattr(archive, '__getitem__'):
        archive = [archive]
    if pretty and pyaml is not None:
        out.write("--- ".join(map(pyaml.dump, archive)))
        return
    return yaml.safe_dump_all(
        archive, out, canonical=True, allow_unicode=True)


class ADFEncoder(json.JSONEncoder):
    def _b64(self, v):
        return {'_base64': b64encode(v)}

    def default(self, o):
        if isinstance(o, (list, dict, str, unicode,
                          int, float, bool, type(None))):
            return super(ADFEncoder, self).default(o)
        if isinstance(o, datetime):
            return format_timestamp(o)
        if o.yaml_tag == Certificate.yaml_tag:
            return {'key': self._b64(o.key)}
        if o.yaml_tag == Auth.yaml_tag:
            return dict([(k, self._b64(v)) for k, v in o.__dict__.iteritems()])
        return o.__dict__


def _as_adf_object(dct):
    """
    May throw parse error for dates
    """
    if '_base64' in dct:
        return b64decode(dct['_base64'])
    if 'key' in dct:
        return Certificate(key=dct['key'])
    if 'md5' in dct:
        return Auth(**dct)
    if 'mode' in dct:
        return Cipher(**dct)
    if 'meta' in dct:
        return Archive(**dct)
    if 'download' in dct or 'upload' in dct or 'local' in dct:
        return Links(**dct)
    if 'created' in dct:
        dct['created'] = parse_timestamp(dct['created'])
        if 'aid' in dct:
            return Signature(**dct)
        else:
            return Meta(**dct)
    return dct


def as_json(docs):
    return ADFEncoder(ensure_ascii=False).encode(docs)


def as_adf(data):
    return json.loads(data, object_hook=_as_adf_object)

json_cert_re = re.compile(r"""
<template><script>var json_certificate='(.*)';</script></template>
""")


def load_archive(f):

    try:
        f = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
    except Exception as e:
        raise InvalidArchiveError(e)
    match = json_cert_re.search(f)
    d = {}
    if match:  # we have a json encoded cert in this file
        try:
            d = as_adf(match.group(1))
        except Exception as e:
            raise InvalidArchiveError(e)
    else:
        for o in load_file(f):
            if isinstance(o, Archive):
                d['archive'] = o
            elif isinstance(o, Auth):
                d['auth'] = o
            elif isinstance(o, Certificate):
                d['cert'] = o
            elif isinstance(o, Links):
                d['links'] = o
            elif isinstance(o, Signature):
                d['signature'] = o
    if 'links' in d and d['links'].download:
        # fix for early client saving archive id in links section
        assert 'signature' not in d
        d['signature'] = Signature(aid=d['links'].download,
                                   uri='http://longaccess.com/a/')
        if d['links'].upload or d['links'].local:
            d['links'].download = None
        else:
            del d['links']
    if 'archive' in d:
        return d
    raise InvalidArchiveError()



from lacore.adf.elements import Links
from lacore.archive.urls import UrlArchiver
from . import Dumper


class UrlDumper(Dumper, UrlArchiver):

    def __init__(self, **kwargs):
        super(UrlDumper, self).__init__(**kwargs)

    def update(self, result):
        super(UrlDumper, self).update(result)
        self.docs['links'] = Links()

    def write(self, data):
        print "dummy writing", len(data)



from functools import update_wrapper, wraps


def expand_args(f):
    @wraps(f)
    def wrap(args=[], kwargs={}):
        return f(*args, **kwargs)
    return wrap


class cached_property(object):
    val = None

    def __init__(self, f):
        self.f = f
        update_wrapper(self, self.f)

    def __get__(self, obj, cls):
        if obj is None:
            return self
        if self.f.__name__ not in obj.__dict__:
            obj.__dict__[self.f.__name__] = self.f(obj)
        return obj.__dict__[self.f.__name__]

    def __set__(self, obj, value):
        obj.__dict__[self.f.__name__] = value


def contains(cls):
    def decorator(func):
        @wraps(func)
        def patched(*args, **kwargs):
            return cls(func(*args, **kwargs))
        return patched
    return decorator


def coroutine(func):
    def start(*args, **kwargs):
        g = func(*args, **kwargs)
        g.next()
        return g
    return start

# vim: et:sw=4:ts=4



from . import CipherBase
from Crypto.Cipher import XOR


class CipherXOR(CipherBase):

    mode = 'xor'
    BLOCKSIZE = 8

    def __init__(self, key, input=None):
        if len(key) != self.BLOCKSIZE:
            raise ValueError("Invalid key")
        self.obj = XOR.new(key)

    def encipher_block(self, block):
        return self.obj.encrypt(block)

    def decipher_block(self, block):
        return self.obj.encrypt(block)

    def flush(self):
        last = super(CipherXOR, self).flush()
        return self.obj.encrypt(last)



from urlparse import urljoin
from lacore import get_client_info
from lacore.decorators import cached_property, contains
from lacore.async import deferred_property, block
from lacore.exceptions import (ApiAuthException, UploadEmptyError,
                               ApiUnavailableException, ApiErrorException)
from lacore.date import parse_timestamp
from lacore.log import getLogger
from twisted.internet import defer
from twisted.python import failure

import json
import treq

from twisted.web.client import HTTPClientFactory
HTTPClientFactory.noisy = True


class TwistedRequestsFactory(object):
    def __init__(self):
        pass

    class TreqSession(object):
        auth = None
        verify = None
        user_agent = None

    class TwistedRequestsSession(object):
        defaults = {
            'persistent': False
        }

        def __init__(self, session):
            self.session = session
            if session.user_agent is not None:
                ua = session.user_agent
            else:
                ua = get_client_info()
            self.defaults.setdefault('headers', {}).update({
                'User-Agent': [str(ua)]})

        @defer.inlineCallbacks
        def get_content(self, r):
            if r.code == 401:
                yield failure.Failure(ApiAuthException())
            if r.code == 404:
                yield failure.Failure(ApiUnavailableException())
            if r.code > 300:
                yield failure.Failure(ApiErrorException(
                    " ".join([str(r.code), str(r.phrase)])))
            r = yield treq.content(r)
            defer.returnValue(json.loads(r))

        def _defaults(self, kwds):
            defaults = {}
            defaults.update(self.defaults)
            defaults.update(kwds)
            return defaults

        @defer.inlineCallbacks
        def maybeauth(self, method, *args, **kwargs):
            try:
                r = yield method(*args, **self._defaults(kwargs))
                r = yield self.get_content(r)
            except ApiAuthException:
                kwargs['auth'] = self.session.auth
                r = yield method(*args, **self._defaults(kwargs))
                r = yield self.get_content(r)
            defer.returnValue(r)

        def get(self, *args, **kwargs):
            return self.maybeauth(treq.get, *args, **kwargs)

        def post(self, *args, **kwargs):
            return self.maybeauth(treq.post, *args, **kwargs)

        def patch(self, *args, **kwargs):
            return self.maybeauth(treq.patch, *args, **kwargs)

    def __call__(self, prefs={}):
        session = self.TreqSession()
        if 'user' in prefs and 'pass' in prefs:
            session.auth = (prefs['user'], prefs['pass'])
        if 'verify' in prefs:
            session.verify = prefs['verify']
        if 'user_agent' in prefs:
            session.user_agent = prefs['user_agent']
        return Api(prefs, self.TwistedRequestsSession(session))


RequestsFactory = TwistedRequestsFactory()


class UploadState(object):
    def __init__(self, archive, size, uri=None, capsule=None, sandbox=False):
        self.archive = archive
        self.size = size
        self.capsule = capsule
        self.sandbox = sandbox
        self.uri = uri


class UploadOperation(object):
    uri = None

    def __init__(self, api, archive, state):
        self.archive = archive
        self.capsule = state.capsule
        self.api = api
        self.uri = state.uri
        self.sandbox = state.sandbox

    @defer.inlineCallbacks
    def start(self):
        endpoints = yield self.api.endpoints
        data = json.dumps({
            'title': self.archive.title,
            'description': self.archive.description or '',
            'capsule': self.capsule['resource_uri'],
            'size': self.archive.meta.size,
            'sandbox': self.sandbox
        })
        r = yield self.api._post(endpoints['upload'], data=data)
        self.uri = urljoin(self.api.url, r['resource_uri'])
        defer.returnValue(self.api.parse_status(r))

    @defer.inlineCallbacks
    def poll(self):
        r = yield self.api._get(self.uri)
        defer.returnValue(self.api.parse_status(r))

    @property
    def status(self):
        if self.uri is None:
            d = self.start()
        else:
            d = self.poll()
        d.addErrback(
            lambda f: getLogger().debug(
                "Exception while getting status", exc_info=True) or f)
        return d

    def finalize(self, auth=None, keys=[]):
        if self.uri is None:
            raise UploadEmptyError(
                reason="Attempt to finalize upload that hasn't started")
        patch = {'status': 'uploaded', 'size': self.archive.meta.size,
                 'parts': len(keys)}
        if auth is not None:
            patch['checksums'] = {}
            if hasattr(auth, 'sha512'):
                patch['checksums']['sha512'] = auth.sha512.encode("hex")
            if hasattr(auth, 'md5'):
                patch['checksums']['md5'] = auth.md5.encode("hex")
        return self.api._patch(self.uri, data=json.dumps(patch))


class Api(object):

    def __init__(self, prefs, session=None):
        self.url = prefs.get('url')
        self.prefs = prefs
        self.session = session

    @deferred_property
    def root(self):
        r = yield self._get(self.url)
        defer.returnValue(r)

    @deferred_property
    def endpoints(self):
        root = yield self.root
        r = yield dict(((n, urljoin(self.url, r['list_endpoint']))
                       for n, r in root.iteritems()))
        defer.returnValue(r)

    @defer.inlineCallbacks
    def _get(self, url):
        r = yield self.session.get(url.encode('utf8'))
        defer.returnValue(r)

    @defer.inlineCallbacks
    def _data(self, url, method, data=None):
        headers = {}
        if data is not None:
            headers['content-type'] = 'application/json'
        r = yield method(
            url.encode('utf8'), headers=headers, data=data)
        defer.returnValue(r)

    def _post(self, url, data=None):
        return self._data(url, self.session.post, data)

    def _patch(self, url, data=None):
        return self._data(url, self.session.patch, data)

    def upload(self, archive, state):
        op = UploadOperation(self, archive, state)
        return op

    def upload_cancel(self, state):
        pass

    @defer.inlineCallbacks
    def get_endpoint(self, name):
        endpoints = yield self.endpoints
        url = endpoints[name]
        r = yield self._get(url)
        defer.returnValue(r)

    @defer.inlineCallbacks
    def async_capsules(self):
        r = yield self.get_endpoint('capsule')
        defer.returnValue(self.capsule_list(r))

    def capsules(self):
        return block(self.async_capsules)()

    @contains(list)
    def object_list(self, os, kwds=None, tstamps=[]):
        for o in os.get('objects'):
            ret = dict([(k, o.get(k, None)) for k in kwds])
            for kw in tstamps:
                ret[kw] = parse_timestamp(ret[kw])
            yield ret

    def capsule_list(self, cs):
        ks = ['title', 'remaining', 'size', 'id', 'resource_uri',
              'created', 'expires']
        return self.object_list(cs, kwds=ks, tstamps=['created', 'expires'])

    @contains(dict)
    def capsule_ids(self, size=None):
        for capsule in self.capsules():
            if size is None or capsule['remaining'] >= size:
                yield (capsule['id'], capsule)

    @defer.inlineCallbacks
    def async_archives(self):
        r = yield self.get_endpoint('archive')
        defer.returnValue(self.archive_list(r))

    def archives(self):
        return block(self.async_archives)()

    def archive_list(self, ars):
        ks = ['capsule', 'description', 'size', 'key', 'resource_uri',
              'created', 'expires', 'title']
        return self.object_list(ars, kwds=ks, tstamps=['created', 'expires'])

    @deferred_property
    def async_account(self):
        r = yield self.get_endpoint('account')
        defer.returnValue(r)

    @cached_property
    @block
    def account(self):
        return self.async_account

    @defer.inlineCallbacks
    def upload_status_async(self, uri):
        r = yield self._get(uri)
        defer.returnValue(self.parse_status(r))

    @block
    def upload_status(self, uri):
        return self.upload_status_async(uri)

    def parse_status(self, rsp):
        if rsp:
            if 'created' in rsp:
                rsp['created'] = parse_timestamp(rsp['created'])
            if 'expires' in rsp:
                rsp['expires'] = parse_timestamp(rsp['expires'])
            if 'archive' in rsp:
                rsp['archive'] = urljoin(self.url, rsp['archive'])
        return rsp
# vim: et:sw=4:ts=4



from boto import set_stream_logger
from lacore.log import setupLogging
from lacore.date import epoch
from binascii import a2b_hex
from contextlib import contextmanager
from tempfile import mkdtemp
from shutil import rmtree


def setup():
    set_stream_logger('boto')
    setupLogging(2)


def makeprefs(factory=None):
    return {
        'api': {
            'user': 'foo',
            'pass': 'bar',
            'url': 'http://baz.com',
            'verify': True,
            'factory': factory
        },
        'upload': {
            'timeout': 1200,
            'bucket': 'lastage',
            'nprocs': None,
        },
        'command': {
            'debug': 0,
            'verbose': False,
            'batch': True,
            'unsafe': False,
            'srm': None
        },
        'gui': {
            'rememberme': False,
            'username': None,
            'password': None,
            'email': None
        }
    }

dummykey = a2b_hex(
    '824aed71bd74c656ed6bdaa19f2a338faedd824d5fd6e96e85b7fac5c6dabe18')

dummycapsule = {
    'title': 'foo',
    'id': 'bar',
    'size': 1230000,
    'remaining': 0,
    'expires': epoch(),
    'created': epoch()
}


@contextmanager
def _temp_home():
    d = mkdtemp()
    yield d
    rmtree(d)



from datetime import datetime

from testtools import TestCase
from struct import unpack
from nose.tools import raises
from mock import Mock


class AdfTest(TestCase):
    def setUp(self):
        super(AdfTest, self).setUp()

    def tearDown(self):
        super(AdfTest, self).tearDown()

    def _archive(self, title='', format='', cipher=''):
        from lacore.adf.elements import Archive, Meta
        return Archive(title, Meta(format, cipher, created='now'))

    def test_archive(self):
        from lacore.adf.persist import make_adf
        self.assertEqual(ADF_TEST_DATA_1,
                         make_adf(self._archive('foo'), out=None,
                                  pretty=False))

    @raises(ValueError)
    def test_archive_invalid(self):
        from lacore.adf.elements import Archive
        Archive('foo', {})

    def test_unicode(self):
        from lacore.adf.persist import make_adf
        self.assertEqual(ADF_TEST_DATA_1,
                         make_adf(self._archive('foo'),
                                  out=None, pretty=False))

    def test_meta(self):
        from lacore.adf.elements import Meta
        from lacore.adf.persist import make_adf
        meta = make_adf(Meta(
            format='zip', cipher='aes-256-ctr', created='now'),
            out=None, pretty=False)
        self.assertEqual(meta, ADF_TEST_DATA_2)

    def test_links(self):
        from lacore.adf.elements import Links
        from lacore.adf.persist import make_adf
        links = make_adf(Links(upload='foo', local='bar'),
                         out=None, pretty=False)
        self.assertEqual(ADF_TEST_DATA_3, links)

    def test_signature(self):
        from lacore.adf.elements import Signature
        from lacore.adf.persist import make_adf
        sig = make_adf(Signature(aid='bar', uri='foo',
                                 created=datetime.utcfromtimestamp(0)),
                       out=None, pretty=False)
        self.assertEqual(ADF_TEST_DATA_4, sig)

    def test_minimal(self):
        from lacore.adf.persist import load_archive

        with open('t/data/archives/minimal.adf') as f:
            docs = load_archive(f)
            self.assertEqual(docs['archive'].meta.cipher, 'aes-256-ctr')
            b = unpack("<LLLLLLLL", docs['cert'].key)
            self.assertEqual(b[0], 1911376514)

    def test_sample(self):
        from lacore.adf.persist import load_archive

        with open('t/data/archives/sample.adf') as f:
            docs = load_archive(f)
            self.assertEqual(docs['archive'].meta.cipher.mode, 'aes-256-ctr')
            b = unpack("<LLLLLLLL", docs['archive'].meta.cipher.input)
            self.assertEqual(b[0], 1911376514)
            b = unpack("<LLLLLLLL", docs['cert'].keys[1].input)
            self.assertEqual(b[0], 1911376514)

    def test_cipher(self):
        from lacore.adf.elements import Cipher
        self.assertRaises(ValueError, Cipher, 'foo', 1)
        self.assertRaises(ValueError, Cipher, 'aes-256-ctr', 'f')
        Cipher('xor', key=1)
        Cipher('xor', key=2, input=Mock())

ADF_TEST_DATA_1 = """---
!archive {
  ? !!str "meta"
  : !meta {
    ? !!str "cipher"
    : !!str "",
    ? !!str "created"
    : !!str "now",
    ? !!str "format"
    : !!str "",
  },
  ? !!str "title"
  : !!str "foo",
}
"""

ADF_TEST_DATA_2 = """---
!meta {
  ? !!str "cipher"
  : !!str "aes-256-ctr",
  ? !!str "created"
  : !!str "now",
  ? !!str "format"
  : !!str "zip",
}
"""

ADF_TEST_DATA_3 = """---
!links {
  ? !!str "local"
  : !!str "bar",
  ? !!str "upload"
  : !!str "foo",
}
"""

ADF_TEST_DATA_4 = """---
!signature {
  ? !!str "aid"
  : !!str "bar",
  ? !!str "created"
  : !!timestamp "1970-01-01 00:00:00",
  ? !!str "expires"
  : !!timestamp "2000-01-01 00:00:00",
  ? !!str "uri"
  : !!str "foo",
}
"""



from testtools import TestCase
from mock import Mock


class AsyncTest(TestCase):
    def setUp(self):
        super(AsyncTest, self).setUp()

    def tearDown(self):
        super(AsyncTest, self).tearDown()

    def test_block_timeout(self):
        from lacore.async import block_timeout, TimeoutError, Failure
        from twisted.internet import defer, reactor

        c = defer.Deferred()
        errback = Mock(side_effect=c.callback)

        @block_timeout(1)
        def long():
            d = defer.Deferred()
            d.addErrback(errback)
            d.addCallback(Mock(side_effect=Exception()))
            reactor.callLater(3, d.callback, None)
            return d

        self.assertRaises(TimeoutError, long)

        @block_timeout(3)
        @defer.inlineCallbacks
        def check():
            try:
                yield c
            except defer.CancelledError:
                self.assertEqual(1, errback.call_count)
                self.assertEqual(1, len(errback.call_args[0]))
                a = errback.call_args[0][0]
                self.assertTrue(isinstance(a, Failure))
                self.assertEqual(a.type, defer.CancelledError)
            else:
                raise Exception("deferred was not cancelled")

        check()

    def test_block(self):
        from lacore.async import block
        from twisted.internet import defer, reactor, task

        d = defer.Deferred()

        @block
        def w():
            r = task.deferLater(reactor, 0.1, d.callback, None)
            self.assertFalse(d.called)
            return r
        w()
        self.assertTrue(d.called)



from __future__ import division

import functools

from abc import ABCMeta, abstractmethod


class CipherBase(object):
    __metaclass__ = ABCMeta

    BLOCKSIZE = 16
    _extra = ''

    @abstractmethod
    def flush(self):
        """ this must be called at the end of the encryption process """
        return self._pad(self._extra)

    def _buffer(self, data, last=False):
        data = self._extra + data
        total = len(data)
        nblocks = total // self.BLOCKSIZE
        if last:
            nblocks -= 1
        pos = 0
        for i in range(nblocks):
            yield data[pos:pos + self.BLOCKSIZE]
            pos += self.BLOCKSIZE
        self._extra = data[pos:]

    def _reduce(self, func, blocks):
        return functools.reduce(lambda r, b: r + func(b), blocks, '')

    def encipher(self, data):
        return self._reduce(self.encipher_block, self._buffer(data))

    def decipher(self, data, last=False):
        ret = self._reduce(self.decipher_block, self._buffer(data, True))
        if last and len(self._extra) > 0:
            ret += self._unpad(self._extra)
            self._extra = ''
        return ret

    @abstractmethod
    def encipher_block(self, block):
        """ pipe a block of data through the cipher """

    @abstractmethod
    def decipher_block(self, block):
        """ pipe a block of data through the cipher """

    def _pad(self, s, bs=None):
        """ pkcs7/rfc5652 padding """
        if bs is None:
            bs = self.BLOCKSIZE
        c = bs - len(s) % bs
        return s + chr(c) * c

    def _unpad(self, p):
        if len(p) != self.BLOCKSIZE:
            raise ValueError("Total input not a multiple of blocksize")
        s = self.decipher_block(p)
        num = ord(s[-1])
        assert num < self.BLOCKSIZE, "plaintext padding error"
        return s[0:-num]



import os
import httpretty

from . import makeprefs, _temp_home
from testtools import TestCase
from unittest import SkipTest
from mock import Mock, patch
from StringIO import StringIO
from urlparse import urlparse


class DumperTest(TestCase):
    def setUp(self):
        super(DumperTest, self).setUp()
        self.prefs = makeprefs()
        self.home = os.path.join('t', 'data')

    def tearDown(self):
        super(DumperTest, self).tearDown()

    @httpretty.activate
    @patch('sys.stdout', new_callable=StringIO)
    def test_dummy_dumper(self, out):
        try:
            import zipstream  # noqa
        except ImportError:
            raise SkipTest("requires python-zipstream")
        httpretty.register_uri(httpretty.GET, "http://foobar/",
                               body='file contents man',
                               content_type="text/plain")
        from lacore.dumper.urls import UrlDumper
        dest = UrlDumper(title='foo')
        cb = Mock()
        list(dest.dump([('http://foobar', 'rel')], cb))
        self.assertTrue('dummy writing' in out.getvalue())

    def test_dumper_no_archive(self):
        from lacore.dumper import Dumper

        class MyDumper(Dumper):
            def write(self, data):
                pass
        dump = MyDumper().dump([], None)
        self.assertRaises(NotImplementedError, list, dump)

    def test_dumper_file(self):
        from lacore.dumper.file import FileDumper
        from lacore.archive.folders import FolderArchiver

        class MyDumper(FileDumper, FolderArchiver):
            pass
        with _temp_home() as tmpdir:
            dest = MyDumper(tmpdir=tmpdir)
            list(dest.dump([os.path.abspath(self.home)], lambda x, y: None))
            self.assertTrue('links' in dest.docs)
            parsed = urlparse(dest.docs['links'].local)
            self.assertTrue(not parsed.scheme or parsed.scheme == 'file')
            self.assertTrue(os.path.exists(parsed.path))

    def test_dumper_file_zip(self):
        from lacore.dumper.file import FileDumper
        from lacore.archive.folders import FolderArchiver
        from lacore.adf.elements import Certificate, Archive, Meta, Cipher

        class MyDumper(FileDumper, FolderArchiver):
            pass
        with _temp_home() as tmpdir:
            docs = {}
            docs['archive'] = Archive(
                'foo', Meta('zip', Cipher('xor', 1)),
                description='bar')
            docs['cert'] = Certificate(key='\0'*8)
            dest = MyDumper(docs=docs, tmpdir=tmpdir)
            list(dest.dump([os.path.abspath(self.home)], lambda x, y: None))
            self.assertTrue('links' in dest.docs)
            parsed = urlparse(dest.docs['links'].local)
            self.assertTrue(not parsed.scheme or parsed.scheme == 'file')
            self.assertTrue(os.path.exists(parsed.path))
            import zipfile
            z = zipfile.ZipFile(parsed.path)
            z.testzip()



import sys


fsenc = sys.getfilesystemencoding() or "UTF-8"


def get_unicode(value):
    if isinstance(value, unicode):
        return value
    try:
        return unicode(value, 'UTF-8')
    except UnicodeDecodeError:
        if fsenc == 'UTF-8':
            raise
        return unicode(value, fsenc)
# vim: et:sw=4:ts=4



import boto
import threading
import tempfile
import os
import sys
from collections import deque

DEFAULT_CHUNK_SIZE = 2**20 * 64  # 64 MB per request
MAX_CACHED = 5


class CachedChunksFile(object):
    """
    Wrap a file object that supports
    Original code from: http://stackoverflow.com/a/9015961/729491
    """

    def __init__(self, file_obj, file_size, chunksize=DEFAULT_CHUNK_SIZE,
                 start=True):
        self._file_obj = file_obj
        self._file_size = file_size
        self._lock = threading.RLock()
        self._load_condition = threading.Condition(self._lock)
        self._load_run = True
        self._loc = 0
        self._exc_info = None
        self._chunk_size = chunksize
        chunk_count = self._file_size // self._chunk_size
        chunk_count += 1 if self._file_size % self._chunk_size else 0
        self._chunks = [None for _ in xrange(chunk_count)]
        self._queue = deque(maxlen=MAX_CACHED)
        self._load_thread = threading.Thread(target=self._load)
        if start:
            self._load_thread.start()

    def _chunk_loc(self):
        """
         Returns (chunk_num, chunk_offset) for a given
         location in the larger file
        """
        return self._loc // self._chunk_size, self._loc % self._chunk_size

    def _load_chunk(self, chunk_num):
        tf = tempfile.TemporaryFile()
        start_idx = chunk_num * self._chunk_size
        self._file_obj.seek(start_idx)
        tf.write(self._file_obj.read(self._chunk_size))
        with self._lock:
            self._chunks[chunk_num] = (tf, tf.tell())  # (tempfile, size)
            if len(self._queue) == self._queue.maxlen:
                last = self._queue[-1]
                self._chunks[last][0].close()
                self._chunks[last] = None
            self._queue.appendleft(chunk_num)
            self._load_condition.notify()

    def _safe_load(self):
        try:
            self._load()
        except Exception:
            self._exc_info = sys.exc_info()

    def _load(self):
        while self._load_run:
            # check current chunk, load if needed
            with self._lock:
                self._load_condition.wait()  # wait till a read comes
                chunk_num, _ = self._chunk_loc()
                chunk_and_size = self._chunks[chunk_num]

            if chunk_and_size is None:
                self._load_chunk(chunk_num)

            # find next empty chunk
            for i in xrange(len(self._chunks)):
                cur_chunk = chunk_num + i
                cur_chunk %= len(self._chunks)  # loop around

                if self._chunks[cur_chunk] is None:
                    self._load_chunk(cur_chunk)
                    break
            else:
                # all done, stop thread
                break

    def seek(self, loc, rel=os.SEEK_SET):
        with self._lock:
            if rel == os.SEEK_CUR:
                self._loc += loc
            elif rel == os.SEEK_SET:
                self._loc = loc
            elif rel == os.SEEK_END:
                self._loc = self._file_size + loc

    def tell(self):
        return self._loc

    def _possibly_raise(self):
        if self._exc_info:
            try:
                raise self._exc_info[0], self._exc_info[1], self._exc_info[2]
            finally:
                self._exc_info = None

    def read(self, bytes_to_read=-1):
        ret = []
        if bytes_to_read < 0:
            bytes_to_read = self._file_size - self._loc
        with self._lock:
            chunk_num, chunk_offset = self._chunk_loc()
            while ((bytes_to_read > 0 or bytes_to_read == -1) and
                   chunk_num < len(self._chunks)):
                while not self._chunks[chunk_num]:
                    self._possibly_raise()
                    self._load_condition.notify()
                    self._load_condition.wait()
                chunk, size = self._chunks[chunk_num]
                cur_chunk_bytes = min(size-chunk_offset,
                                      bytes_to_read, size)
                chunk.seek(chunk_offset, os.SEEK_SET)
                data = chunk.read(cur_chunk_bytes)
                bytes_to_read -= len(data)
                ret.append(data)
                chunk_num += 1
                chunk_offset = 0
            ret = ''.join(ret)
            self._loc += len(ret)
        return ret

    def start(self):
        if self._load_thread.isAlive():
            return
        self._load_thread.start()

    def stop(self):
        self._load_run = False
        with self._lock:
            self._load_condition.notify()

    def __enter__(self):
        self._load_run = True
        self.start()
        return self

    def close(self):
        self.stop()
        self._load_thread.join(timeout=3)
        assert not self._load_thread.isAlive()

    def __exit__(self, type, value, traceback):
        self.close()


class S3RangeReader:
    def __init__(self, key_obj):
        self._key_obj = key_obj
        self.size = self._key_obj.size
        self._pos = 0

    def __len__(self):
        return self.size

    def seek(self, pos, rel=os.SEEK_SET):
        if rel == os.SEEK_CUR:
            self._pos += pos
        elif rel == os.SEEK_SET:
            self._pos = pos
        elif rel == os.SEEK_END:
            self._pos = self.size + pos

    def read(self, bytes=-1):
        if bytes == 0 or self._pos >= self.size:
            return ''
        else:
            if bytes == -1:
                bytes = self.size
            # S3 ranges are closed ranges: [start,end]
            headers = {'Range': 'bytes=%s-%s' % (
                self._pos, self._pos + bytes - 1)}
            return self._key_obj.get_contents_as_string(headers=headers)


class S3BackedFile(CachedChunksFile):
    def __init__(self, key, **kwargs):
        super(S3BackedFile, self).__init__(
            S3RangeReader(key), key.size, **kwargs)

if __name__ == '__main__':
    key = boto.connect_s3().get_bucket('lastage').get_key('foobarbig2')
    with S3BackedFile(key) as bf:  # download starts by default
        import zipfile
        with zipfile.ZipFile(bf) as z:
            for zi in z.infolist():
                print zi.filename
                z.extract(zi, '/tmp/')



# -*- coding: utf-8 -*-
from testtools import TestCase
# from mock import Mock
from binascii import a2b_hex, b2a_hex

#  NIST test vectors (last one is ours, uses padding)
KEY = '603deb1015ca71be2b73aef0857d77811f352c073b6108d72d9810a30914dff4'
IV = 'f0f1f2f3f4f5f6f7f8f9fafbfcfdfeff'
VECTORS = (
    ('6bc1bee22e409f96e93d7e117393172a', '601ec313775789a5b7a7f504bbf3d228'),
    ('ae2d8a571e03ac9c9eb76fac45af8e51', 'f443e3ca4d62b59aca84e990cacaf5c5'),
    ('30c81c46a35ce411e5fbc1191a0a52ef', '2b0930daa23de94ce87017ba2d84988d'),
    ('f69f2445df4f9b17ad2b417be66c3710', 'dfc9c58db67aada613c2dd08457941a6'),
    ('2ab7a7bc7f73420932',               'a1c0585ca60f4b9be5f0091beec8c4b0'))


class CipherAESTest(TestCase):

    def setUp(self):
        super(CipherAESTest, self).setUp()

    def tearDown(self):
        super(CipherAESTest, self).tearDown()

    def _makeit(self, *args, **kwargs):
        from lacore.cipher.aes import CipherAES
        return CipherAES(*args, **kwargs)

    def test_encipher_invalid(self):
        self.assertRaises(ValueError, self._makeit, None, None)

    def test_encipher(self):
        cipher = self._makeit(a2b_hex(KEY),
                              a2b_hex(IV))
        for p, c in VECTORS:
            if len(a2b_hex(p)) < cipher.BLOCKSIZE:
                self.assertEqual(c, b2a_hex(cipher.encipher(a2b_hex(p))) +
                                 b2a_hex(cipher.flush()))
            else:
                self.assertEqual(c, b2a_hex(cipher.encipher(a2b_hex(p))))

    def test_encipher_all(self):
        cipher = self._makeit(a2b_hex(KEY),
                              a2b_hex(IV))
        pt = "".join([v[0] for v in VECTORS])
        ct = "".join([v[1] for v in VECTORS])
        self.assertEqual(ct, b2a_hex(cipher.encipher(a2b_hex(pt))) +
                         b2a_hex(cipher.flush()))

    def test_encipher_flush(self):
        cipher = self._makeit(a2b_hex(KEY),
                              a2b_hex(IV))
        pt = VECTORS[0][0]
        ct = VECTORS[0][1]
        self.assertEqual(ct, b2a_hex(cipher.encipher(a2b_hex(pt))))

    def test_flush(self):
        cipher = self._makeit("0"*32)
        # padding = a2b_hex("10"*16)
        # ct = b2a_hex(self._makeit(Mock(key="0"*32)).encipher_block(padding))
        ct = '2b9cf21e2f5611a68ab213aa44972cf0'
        self.assertEqual(ct, b2a_hex(cipher.flush()))

    def test_decipher(self):
        cipher = self._makeit(a2b_hex(KEY),
                              a2b_hex(IV))
        for i in range(len(VECTORS)+1):
            p = c = ''
            last = i == len(VECTORS)
            if i:
                p = VECTORS[i-1][0]
            if not last:
                c = VECTORS[i][1]
            self.assertEqual(p, b2a_hex(cipher.decipher(a2b_hex(c), last)))

    def test_decipher_all(self):
        cipher = self._makeit(a2b_hex(KEY),
                              a2b_hex(IV))
        pt = "".join([v[0] for v in VECTORS])
        ct = "".join([v[1] for v in VECTORS])
        self.assertEqual(pt, b2a_hex(cipher.decipher(a2b_hex(ct), True)))



from io import DEFAULT_BUFFER_SIZE
from . import CipherBase


class CipherNone(CipherBase):
    BLOCKSIZE = DEFAULT_BUFFER_SIZE

    mode = 'none'

    def __init__(self, key, input=None):
        pass

    def encipher_block(self, block):
        return block

    def decipher_block(self, block):
        return block

    def flush(self):
        return super(CipherNone, self).flush()

    def _pad(self, s):
        return s

    def _unpad(self, s):
        return s



from lacore.date import parse_timestamp, epoch


def creation(docs):
    """ return signature creation, or if not available, the archive creation.
        For sorting purposes. Invalid timestamps sort to the start of the epoch
    """
    created = docs['archive'].meta.created
    sig = docs.get('signature')
    if sig and sig.created:
        created = sig.created
    tstamp = parse_timestamp(created)
    if not tstamp:
        tstamp = epoch()
    return tstamp


def archive_size(archive):
    size = ""
    if archive.meta.size:
        kib = archive.meta.size // 1024
        mib = kib // 1024
        gib = mib // 1024

        if not kib:
            size = "{} B".format(archive.meta.size)
        elif not mib:
            size = "{} KB".format(kib)
        elif not gib:
            size = "{} MB".format(mib)
        else:
            size = "{} GB".format(gib)
    return size



# -*- coding: utf-8 -*-
import os
import sys
import operator
import zipfile
import httpretty

from testtools import TestCase
from unittest import SkipTest
from . import makeprefs, _temp_home
from tempfile import NamedTemporaryFile
from mock import MagicMock, Mock, patch
from itertools import islice, izip, count


class ArchiveTest(TestCase):
    def setUp(self):
        super(ArchiveTest, self).setUp()
        self.prefs = makeprefs()
        self.data = os.path.join('t', 'data')

    def tearDown(self):
        super(ArchiveTest, self).tearDown()

    def test_restore(self):
        from lacore.archive import restore_archive
        from lacore.adf.persist import load_archive
        with _temp_home() as tmpdir:
            cb = Mock()
            cfname = os.path.join(
                self.data, '2013-10-22-foobar.zip.crypt')
            with open(os.path.join(self.data,
                                   'certs',
                                   '2013-10-13-foobar.adf')) as cf:
                cert = load_archive(cf)
            restore_archive(
                cert['archive'], cfname, cert['cert'], tmpdir, tmpdir, cb)
            self.assertEqual(1, cb.call_count)
            self.assertTrue(os.path.exists(
                os.path.join(tmpdir, 'xtQz6ziJ.sh.part')))

    def test_walk_folders(self):
        from lacore.archive.folders import walk_folders
        sample = None
        files = [os.path.abspath(self.data),
                 os.path.abspath(
                     os.path.join('t', 'data', 'longaccess-74-5N93.html'))]
        for p, r in walk_folders(files):
            self.assertEqual(unicode, type(r))
            if r.endswith('sample.adf'):
                sample = r
        self.assertEqual("data/archives/sample.adf", sample)

    def test_folder_args(self):
        from lacore.archive.folders import FolderArchiver

        cb = Mock()
        ar = FolderArchiver()
        for a, k in ar.args([self.data], cb):
            self.assertTrue('arcname' in k)
            self.assertEqual(str, type(k['arcname']))

        self.assertTrue(u'data/archives/sample.adf' in
                        map(operator.itemgetter(1),
                            map(operator.itemgetter(0),
                                cb.call_args_list)))

    def test_folder_args_exception(self):
        from lacore.archive.folders import FolderArchiver

        cb = Mock(side_effect=Exception())
        ar = FolderArchiver()
        args = ar.args([self.data], cb)
        e = self.assertRaises(Exception, list, args)
        self.assertTrue(hasattr(e, 'filename'))

    @httpretty.activate
    def test_url_args(self):
        s = (self.getUniqueString() for _ in xrange(100))
        mkpath = lambda n: "/".join(islice(s, n))
        paths = [(mkpath(n),                              # relative path
                  next(s),                                # content
                  "http://" + next(s) + "/" + mkpath(n))  # url
                 for n in xrange(5)]
        for p, c, u in paths:
            httpretty.register_uri(httpretty.GET, u, body=c)
        from lacore.archive.urls import UrlArchiver
        cb = Mock()
        ar = list(UrlArchiver().args([(u, p) for p, c, u in paths], cb))
        for p, a, c in zip(paths, ar, cb.call_args_list):
            self.assertTrue('arcname' in a[1])
            self.assertEqual(str, type(a[1]['arcname']))
            self.assertEqual(p[0], a[1]['arcname'])
            self.assertEqual(p[0], c[0][1])
            with a[0][0] as contents:
                self.assertEqual(p[1], contents.read())

    @httpretty.activate
    def test_url_args_deprecated(self):
        s = (self.getUniqueString() for _ in xrange(100))
        fname = next(s)+'.txt'
        path = next(s)
        url = 'http://{}/{}'.format(next(s), fname)
        httpretty.register_uri(httpretty.GET, url, body='foo')
        from lacore.archive.urls import UrlArchiver
        ar = UrlArchiver().args([url, (url, path)], Mock())
        f, a = next(ar)
        self.assertEqual(fname, a['arcname'])
        with f[0] as contents:
            self.assertEqual('foo', contents.read())
        f, a = next(ar)
        self.assertEqual(path, a['arcname'])
        with f[0] as contents:
            self.assertEqual('foo', contents.read())

    @httpretty.activate
    def test_url_args_progress(self):
        s = (self.getUniqueString() for _ in xrange(100))
        paths = list(islice(s, 10))
        url = 'http://{}/{}'.format(next(s), next(s))
        body = self.getUniqueString()
        httpretty.register_uri(httpretty.GET, url, body=body)
        from lacore.archive.urls import UrlArchiver
        ar = UrlArchiver()
        args = ar.args([(url, p) for p in paths], Mock())
        for n, a, p in izip(count(), args, paths):
            self.assertEqual(n+1, len(ar.processed))
            self.assertEqual(p, a[1]['arcname'])
            self.assertEqual(0, ar.processed[p])
            self.assertEqual(body, a[0][0].read())
            self.assertEqual(len(body), ar.processed[p])

    def test_zip_archiver_tempfile(self):
        from lacore.archive.zip import ZipArchiver
        ar = ZipArchiver()
        foo = MagicMock()
        with ar._temp_file(foo) as f:
            f.write("BAR")
        foo.__enter__.assert_called()
        foo.__enter__.return_value.write.assert_called_with("BAR")

    def test_zip_archiver_args(self):
        from lacore.archive.zip import ZipArchiver
        ar = ZipArchiver()
        cb = Mock()
        self.assertEqual([], list(ar.args([], cb)))
        cb.assert_not_called()
        self.assertEqual([(('bar',), {})], list(ar.args(['bar'], cb)))
        cb.assert_called_with('bar', 'bar')

    def test_zip_archiver_no_zipstream(self):
        orig_import = __import__

        def import_mock(name, *args):
            if name == 'zipstream':
                raise ImportError()
            return orig_import(name, *args)
        with patch('__builtin__.__import__', side_effect=import_mock):
            if 'lacore.archive.zip' in sys.modules:
                del sys.modules['lacore.archive.zip']
            import lacore.archive.zip as azip
            with NamedTemporaryFile() as f:
                dst = MagicMock()
                dst.__enter__.return_value = f
                ar = azip.ZipArchiver()
                ar._temp_file = Mock(return_value=dst)
                list(ar.archive([os.path.abspath(os.path.join(
                    't', 'data', 'longaccess-74-5N93.html'))], dst, None))
                dst.__enter__.assert_called()
                ar._temp_file.assert_called()
                z = zipfile.ZipFile(f)
                z.testzip()

    def test_zip_archiver_zipstream(self):
        try:
            import zipstream  # noqa
        except ImportError:
            raise SkipTest("requires python-zipstream")
        if 'lacore.archive.zip' in sys.modules:
            del sys.modules['lacore.archive.zip']
        from lacore.archive.zip import ZipArchiver

        with NamedTemporaryFile() as f:
            dst = MagicMock()
            dst.__enter__.return_value = f
            ar = ZipArchiver()
            ar._temp_file = Mock(return_value=dst)
            list(ar.archive([os.path.abspath(os.path.join(
                't', 'data', 'longaccess-74-5N93.html'))], dst, None))
            dst.__enter__.assert_called()
            ar._temp_file.assert_not_called()
            z = zipfile.ZipFile(f)
            z.testzip()



from lacore.source.stream import StreamSource
from lacore.auth import MyHashObj
from lacore.decorators import coroutine
from lacore.adf.elements import Certificate, Archive, Meta, Cipher
from lacore.exceptions import InvalidArchiveError
from lacore.cipher import get_cipher
from lacore.crypt import CryptIO
from abc import ABCMeta, abstractmethod
from shutil import copyfileobj


class Dumper(object):
    __metaclass__ = ABCMeta

    def __init__(self, docs={}, title='', description=None, fmt='zip',
                 hashf='sha512', chunk=100, **kwargs):
        super(Dumper, self).__init__(**kwargs)
        self.hashobj = MyHashObj(hashf)
        self.chunk = chunk
        self.docs = docs
        if 'archive' not in self.docs:
            self.docs['archive'] = Archive(
                title, Meta(fmt, Cipher('aes-256-ctr', 1)),
                description=description)
        if 'cert' not in self.docs:
            self.docs['cert'] = Certificate()

    def update(self, result):
        self.docs.update(result)

    @abstractmethod
    def write(self, data):
        """ write a piece of data to the storage """

    @coroutine
    def uploader(self):
        while True:
            data = yield
            self.write(data)

    @coroutine
    def end(self):
        self.docs['archive'].meta.size = yield
        self.update({'auth': self.hashobj.auth()})

    @coroutine
    def verify(self):
        yield
        if not self.hashobj.verify(self.docs['auth']):
            raise InvalidArchiveError("archive verification failed")
        self.update({})

    def dump(self, items, cb=None):
        if not hasattr(self, 'archive'):
            raise NotImplementedError(
                "you need to inherit from an archiver class")
        with self:
            with StreamSource(self.end(), self.uploader(), self.chunk) as dest:
                cipher = get_cipher(self.docs['archive'], self.docs['cert'])
                with CryptIO(dest, cipher, hashobj=self.hashobj) as fdst:
                    for result in self.archive(items, fdst, cb):
                        yield result

    def dedump(self, infile, chunk=1024):
        cipher = get_cipher(self.docs['archive'], self.docs['cert'])
        with self:
            with StreamSource(self.verify(),
                              self.uploader(), self.chunk) as dest:
                with CryptIO(infile, cipher,
                             hashobj=self.hashobj, mode='rb') as cf:
                    copyfileobj(cf, dest, chunk)

    def __enter__(self):
        return self

    def __exit__(self, eType, eValue, eTrace):
        return



import sys
from io import BufferedIOBase


class StreamSource(BufferedIOBase):

    def __init__(self, dst, out, chunk=100):
        self.mode = 'w'
        self.end = dst
        self.out = out
        self.size = 0
        self.dst = self.sink()
        self.dst.send(None)
        self.buf = ""
        self.bufsize = 0
        self.chunk = chunk

    def sink(self):
        try:
            while True:
                data = yield
                if data is None:
                    break
                else:
                    self.out.send(data)
            self.end.send(self.size)
        except Exception as e:
            self.end.throw(e)
            raise Exception("Generator didn't stop after throw")

    def _raise_if_closed(self):
        if self.closed:
            raise ValueError("I/O operation on closed file.")

    def write(self, data):
        self._raise_if_closed()
        # Convert data type if called by io.BufferedWriter.
        if isinstance(data, memoryview):
            data = data.tobytes()
        sz = len(data)
        if sz > 0:
            newbuf = self.buf + data
            newbufsize = self.bufsize + sz
            while newbufsize >= self.chunk:
                self.size += self.chunk
                chunk = newbuf[:self.chunk]
                newbuf = newbuf[self.chunk:]
                newbufsize = newbufsize - self.chunk
                self.dst.send(chunk)
            self.buf = newbuf
            self.bufsize = newbufsize
        return sz

    def flush(self):
        self._raise_if_closed()
        if self.bufsize > 0:
            self.size += self.bufsize
            self.bufsize = 0
            self.dst.send(self.buf)
            self.buf = ''

    def close(self):
        if self.closed:
            return
        super(StreamSource, self).close()
        try:
            self.dst.send(None)
        except StopIteration:
            return
        else:
            raise Exception("Generator didn't stop")

    def readable(self):
        return False

    def writable(self):
        return True

    def __exit__(self, eType, eValue, eTrace):
        if eType is None:
            super(StreamSource, self).__exit__(eType, eValue, eTrace)
        else:
            try:
                self.dst.throw(eType, eValue, eTrace)
            except:
                if sys.exc_info()[1] is not eValue:
                    raise
            finally:
                self.bufsize = 0  # no point in flushing data anymore
                self.close()



import os
from testtools import TestCase
from lacore.decorators import coroutine


class StreamSourceTest(TestCase):
    def setUp(self):
        super(StreamSourceTest, self).setUp()
        self.home = os.path.join('t', 'data', 'home')
        self.testfile = os.path.join('t', 'data', 'longaccess-74-5N93.html')

    def tearDown(self):
        super(StreamSourceTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacore.source.stream import StreamSource
        return StreamSource(*args,  **kw)

    def test_constructor_none(self):
        self._makeit(None, None)

    def test_constructor_chunk(self):
        f = self._makeit(None, None, chunk=123)
        self.assertEqual(123, f.chunk)
        self.assertEqual(False, f.readable())
        self.assertEqual(True, f.writable())

    @coroutine
    def coro(self, *args):
        for v in args:
            v2 = yield
            yield self.assertEqual(v, v2)

    def test_write(self):
        vs = ["x"*50, "x"*50]
        f = self._makeit(self.coro(100), self.coro("x"*100))
        for v in vs:
            self.assertEqual(len(v), f.write(v))
        f.close()

    def test_write_context(self):
        with self._makeit(self.coro(100), self.coro("x"*100)) as f:
            self.assertEqual(50, f.write("x"*50))
            self.assertEqual(50, f.write("x"*50))

    def test_write_context2(self):
        with self._makeit(self.coro(0), self.coro("x"*100)) as f:
            f.close()

    def test_write_context3(self):
        with self._makeit(self.coro(0), self.coro("x"*100)) as f:
            f.close()
            self.assertRaises(ValueError, f.write, "BAR")

    def test_write_context4(self):
        with self._makeit(self.coro(90), self.coro("x"*90)) as f:
            self.assertEqual(50, f.write("x"*50))
            self.assertEqual(40, f.write("x"*40))

    def test_write_context5(self):
        with self._makeit(self.coro(90), self.coro("x"*90), chunk=90) as f:
            self.assertEqual(50, f.write("x"*50))
            self.assertEqual(40, f.write("x"*40))

    def test_write_context6(self):
        f = self._makeit(self.coro(90), self.coro("x"*90), chunk=90)
        f.dst = self.coro(None)
        e = self.assertRaises(Exception, f.close)
        self.assertEqual("Generator didn't stop", str(e))

    def test_write_context7(self):

        def mustraise():
            with self._makeit(self.coro(90), self.coro("x"*90)):
                raise Exception('lala')
        e = self.assertRaises(Exception, mustraise)
        self.assertEqual("lala", str(e))

        def catch():
            try:
                foo = yield
            except:
                foo = ''
            yield foo

        def mustraise2():
            with self._makeit(self.coro(90), self.coro("x"*90)) as f:
                f.end = catch()
                f.end.send(None)
                raise Exception('lala')
        e = self.assertRaises(Exception, mustraise2)
        self.assertEqual("Generator didn't stop after throw", str(e))



class BaseAppException(Exception):
    msg = "error"

    def __init__(self, exc=None, *args, **kwargs):
        super(BaseAppException, self).__init__(self.msg, *args, **kwargs)
        self.exc = exc

    def __str__(self):
        return self.msg


class ApiErrorException(BaseAppException):
    msg = "the server couldn't fulfill your request"


class ApiUnavailableException(BaseAppException):
    msg = "resource not found"


class ApiAuthException(BaseAppException):
    msg = "authentication failed"

    def __init__(self, username=None, *args, **kwargs):
        super(ApiAuthException, self).__init__(*args, **kwargs)
        if username:
            self.msg = "authentication as '{}' failed".format(username)


class UploadError(BaseAppException):
    msg = "upload failed"

    def __init__(self, reason=None, *args, **kwargs):
        if reason is not None:
            self.msg += ": {}".format(reason)
        super(BaseAppException, self).__init__(self.msg, *args, **kwargs)


class UploadEmptyError(UploadError):
    pass


class UploadInvalidError(UploadError):
    pass


class InvalidArchiveError(BaseAppException):
    msg = "invalid archive"


class DecryptionError(BaseAppException):
    msg = "Error decrypting file"

    def __init__(self, reason=None):
        super(DecryptionError, self).__init__()
        self.reason = reason



# -*- coding: utf-8 -*-
from nose.tools import raises
from testtools import TestCase
from binascii import a2b_hex, b2a_hex
from mock import Mock, patch


class CipherTest(TestCase):
    def setUp(self):
        super(CipherTest, self).setUp()

    def tearDown(self):
        super(CipherTest, self).tearDown()

    @raises(TypeError)
    def test_cipher_base(self):
        from lacore.cipher import CipherBase
        CipherBase()

    @raises(TypeError)
    def test_cipher_base_inherit(self):
        from lacore.cipher import CipherBase

        class Foo(CipherBase):
            mode = 'foo'
        Foo()

    def _makeit(self, *args, **kwargs):
        from lacore.cipher import CipherBase

        class Foo(CipherBase):
            mode = 'foo'

            def flush(self):
                return super(Foo, self).flush()

            def encipher_block(self, block):
                return block

            def decipher_block(self, block):
                return block
        return Foo(*args, **kwargs)

    def test_cipher(self):
        cipher = self._makeit()
        self.assertFalse(cipher.encipher("0"*8))
        self.assertEqual("0"*16, cipher.encipher("0"*20))
        self.assertEqual("0"*16, cipher.encipher("0000"))

    def test_decipher(self):
        c1 = self._makeit()
        c2 = self._makeit()
        self.assertFalse(c2.decipher(c1.encipher("0"*8)))
        self.assertFalse(c2.decipher(c1.encipher("0"*20)))
        self.assertEqual("0"*32, c2.decipher(c1.encipher("0"*24)))
        self.assertEqual("0"*20, c2.decipher(c1.flush(), True))

    def test_decipher_remaining(self):
        cipher = self._makeit()
        self.assertFalse(cipher.decipher("0"*28))
        self.assertRaises(ValueError, cipher.decipher, "000000", True)

    def test_xor(self):
        from lacore.cipher.xor import CipherXOR
        self.assertRaises(ValueError, CipherXOR, '1')
        cipher = CipherXOR(a2b_hex('ff00ff00ff00ff00'))
        self.assertEqual("5aa55aa55aa55aa5"*2+"f708f708f708f708", b2a_hex(
            cipher.encipher(a2b_hex('a5a5a5a5a5a5a5a5'*2))+cipher.flush()))
        self.assertEqual("fb04fb04fb04fb04", b2a_hex(cipher.encipher(
            a2b_hex('04040404'))+cipher.flush()))

    def test_xor_dec(self):
        from lacore.cipher.xor import CipherXOR
        c1 = CipherXOR(a2b_hex('ff00ff00ff00ff00'))
        c2 = CipherXOR(a2b_hex('ff00ff00ff00ff00'))
        self.assertFalse(c2.decipher(c1.encipher("a5a5")))  # 4
        self.assertEqual("a5"*8, c2.decipher(c1.encipher("a5"*12)))  # 24
        self.assertEqual("a5"*6, c2.decipher(c1.flush(), True))

    def test_pad(self):
        cipher = self._makeit()
        padded = cipher._pad(a2b_hex("00"*10))
        self.assertEqual("00"*10+"06"*6, b2a_hex(padded))
        self.assertEqual("00"*10, b2a_hex(cipher._unpad(padded)))

    def test_get_cipher(self):
        from lacore.adf.elements import Archive, Meta, Cipher
        from lacore.cipher import get_cipher
        with patch('lacore.cipher.xor.CipherXOR.__init__') as xorinit:
            xorinit.return_value = None
            archive = Archive('foo', Meta('zip', Cipher('xor', 1, 'bar')))
            get_cipher(archive, Mock(key='baz'))
            xorinit.assert_called_with('baz', 'bar')
            cert = Mock(keys=['baz'])
            del cert.key
            get_cipher(archive, cert)
            xorinit.assert_called_with('baz', 'bar')
            archive = Archive('foo', Meta('zip', Cipher('xor', 2, 'bar')))
            cert.keys.append('spam')
            get_cipher(archive, cert)
            archive = Archive('foo', Meta('zip', 'xor'))
            get_cipher(archive, cert)



import os
from testtools import TestCase
from mock import Mock, patch
from hashlib import md5


class ChunkedFileTest(TestCase):
    def setUp(self):
        super(ChunkedFileTest, self).setUp()
        self.home = os.path.join('t', 'data', 'home')
        self.testfile = os.path.join('t', 'data', 'longaccess-74-5N93.html')

    def tearDown(self):
        super(ChunkedFileTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacore.source.chunked import ChunkedFile
        return ChunkedFile(*args,  **kw)

    def test_constructor_nopath(self):
        e = self.assertRaises(IOError, self._makeit, 'foobar')
        self.assertEqual("File foobar not found", str(e))

    def test_constructor_not_file(self):
        f = self._makeit(os.devnull)
        self.assertFalse(f.isfile)

    def test_constructor_devnull(self):
        f = self._makeit(os.devnull)
        self.assertFalse(f.isfile)

    def test_constructor_file(self):
        f = self._makeit(self.testfile)
        self.assertTrue(f.isfile)

    def test_constructor_file_too_big_skip(self):
        self.assertRaises(AssertionError, self._makeit, self.testfile,
                          skip=50000000)

    def test_constructor_file_one_chunk(self):
        f = self._makeit(self.testfile, chunk=50000000000)
        self.assertEqual(1, f.chunks)

    def test_constructor_file_one_chunk2(self):
        f = self._makeit(os.path.join('t', 'data', 'empty.txt'))
        self.assertEqual(1, f.chunks)

    def test_chunkstart(self):
        fsz = os.path.getsize(self.testfile)
        for sz in [10, 15]:
            f = self._makeit(self.testfile, chunk=sz)
            last = int(fsz/sz)
            for chunk in range(last):
                if chunk > last:
                    self.assertRaises(ValueError, f.chunkstart, chunk)
                else:
                    self.assertEqual(chunk * sz, f.chunkstart(chunk))

    def test_chunksize(self):
        fsz = os.path.getsize(self.testfile)
        for sz in [10, 15]:
            f = self._makeit(self.testfile, chunk=sz)
            last = int(fsz/sz)
            lastsz = fsz % sz
            if lastsz > 0:
                last += 1
            else:
                lastsz = sz
            for chunk in range(last):
                if chunk == last - 1:
                    self.assertEqual(lastsz, f.chunksize(chunk))
                else:
                    self.assertEqual(sz, f.chunksize(chunk))
            self.assertRaises(ValueError, f.chunksize, last + 1)

    @patch('lacore.source.chunked.compute_md5')
    def test_filehash(self, compute_md5):
        from lacore.source.chunked import SavedPart

        compute_md5.return_value = ('MD5', 'B64MD5', 123)
        src = Mock()
        src.chunksize.return_value = 123
        f = SavedPart(src, 1, self.testfile)
        self.assertEqual(2, len(f.hash))
        self.assertEqual('MD5', f.hash[0])
        self.assertEqual('B64MD5', f.hash[1])
        self.assertEqual(123, f.bytes)
        src.chunksize.assert_called_with(1)

    def test_filepart(self):
        src = Mock()
        src.path = self.testfile
        src.chunkstart.return_value = 2
        src.chunksize.return_value = 7
        from lacore.source.chunked import FilePart
        f = FilePart(src, 1)
        src.chunksize.assert_called_with(1)
        src.chunkstart.assert_called_with(1)
        self.assertEqual('DOCTYPE', f.read())
        m = md5()
        m.update('DOCTYPE')
        self.assertEqual(m.hexdigest(), f.hash[0])



from io import BufferedIOBase
from errno import EBADF
from lacore.exceptions import DecryptionError

READ, WRITE = 1, 2


class CryptIO(BufferedIOBase):

    def __init__(self, fileobj, cipher, mode=None, hashobj=None):
        if mode:
            mode = mode.replace('U', '')
        if mode and 'b' not in mode:
            mode += 'b'
        if mode is None:
            if hasattr(fileobj, 'mode'):
                mode = fileobj.mode
            else:
                mode = 'rb'
        _m = mode[0:1]
        if _m == 'r':
            self.mode = READ
            # From GzipFile:
            # Buffer data read from underlying file. extrastart is offset in
            # stream where buffer starts. extrasize is number of bytes
            # remaining in buffer from current stream position.
            self.extrabuf = ""
            self.extrasize = 0
            self.extrastart = 0
        elif _m == 'w' or _m == 'a':
            self.mode = WRITE
        else:
            raise IOError("Mode '{}' not supported".format(mode))

        if self.mode == WRITE and not hasattr(fileobj, 'write'):
            raise ValueError("Invalid file object")

        if self.mode == WRITE and not hasattr(cipher, 'encipher'):
            raise ValueError("Invalid cipher object")

        if self.mode == READ and not hasattr(cipher, 'decipher'):
            raise ValueError("Invalid cipher object")

        if self.mode == READ:
            self.fileobj = iter(fileobj)
        else:
            self.fileobj = fileobj
        self.cipher = cipher

        self.offset = 0
        self.size = 0
        self.hashobj = hashobj

    def _raise_if_closed(self):
        if self.closed:
            raise ValueError("I/O operation on closed file.")

    def write(self, data):
        self._raise_if_closed()
        if self.mode != WRITE:
            raise IOError(EBADF, "write() on read-only CryptIO object")

        # Convert data type if called by io.BufferedWriter.
        if isinstance(data, memoryview):
            data = data.tobytes()

        sz = len(data)
        if sz > 0:
            self.size += sz
            towrite = self.cipher.encipher(data)
            self.fileobj.write(towrite)
            if self.hashobj:
                self.hashobj.update(towrite)
            self.offset += sz
        return sz

    def read(self, size=-1):
        self._raise_if_closed()
        if self.mode != READ:
            raise IOError(EBADF, "read() on write-only CryptIO object")

        _finished = lambda: size >= 0 and size <= self.extrasize

        def _add(data):  # put data into read buffer
            offset = self.offset - self.extrastart
            self.extrabuf = self.extrabuf[offset:] + data
            self.extrasize += len(data)
            self.extrastart = self.offset
            self.size += len(data)

        while not _finished():
            try:
                buf = next(self.fileobj)
                if len(buf) > 0:
                    if self.hashobj:
                        self.hashobj.update(buf)
                    _add(self.cipher.decipher(buf))
            except StopIteration:
                _add(self.cipher.decipher('', True))
                if not _finished():
                    size = self.extrasize
                break

        start = self.offset - self.extrastart
        stop = start + size

        self.extrasize -= size
        self.offset += size

        return self.extrabuf[start:stop]

    def flush(self):
        self._raise_if_closed()
        if self.mode == WRITE:
            towrite = self.cipher.flush()
            if self.hashobj:
                self.hashobj.update(towrite)
            self.fileobj.write(towrite)
            self.fileobj.flush()

    def close(self):
        if self.closed:
            return
        super(CryptIO, self).close()
        self.fileobj = None

    def readable(self):
        return self.mode == READ

    def writable(self):
        return self.mode == WRITE

    def __exit__(self, eType, eValue, eTrace):
        super(CryptIO, self).__exit__(eType, eValue, eTrace)
        if eType and self.mode == READ:
            raise DecryptionError(eValue)



import sys
import os

from .zip import ZipArchiver
from lacore.enc import get_unicode
from itertools import imap


def walk_folders(folders):
    for folder in folders:
        if not os.path.isdir(folder):
            yield (folder, get_unicode(os.path.basename(folder)))
        else:
            for root, _, fs in os.walk(folder):
                for f in fs:
                    path = os.path.join(root, f)
                    strip = os.path.dirname(folder)
                    rel = os.path.relpath(path, strip)
                    yield (path, get_unicode(rel))


class FolderArchiver(ZipArchiver):
    def args(self, items, cb):
        if sys.platform.startswith('win'):
            # windows has unicode file system api
            items = imap(unicode, items)
        for path, rel in walk_folders(imap(os.path.abspath, items)):
            try:
                cb(path, rel)
                yield ((path,), {'arcname': rel.encode('utf-8')})
            except Exception as e:
                if not hasattr(e, 'filename'):
                    setattr(e, 'filename', path)
                raise



from testtools import TestCase
from testtools.deferredruntest import assert_fails_with
from mock import patch, Mock, DEFAULT
from twisted.internet import defer
from lacore.exceptions import (ApiAuthException, ApiUnavailableException,
                               ApiErrorException)
from lacore.t.test_api import LA_ENDPOINTS_RESPONSE, LA_CAPSULES_RESPONSE
from itertools import imap
from . import makeprefs
import crochet
import json


class ApiAsyncTest(TestCase):

    @classmethod
    def setup_class(cls):
        crochet.setup()

    def _makeit(self, prefs={}):
        from lacore.api import RequestsFactory
        with patch('lacore.api.Api') as api:
            RequestsFactory(prefs)
            args, _ = api.call_args
            return args[1]

    @crochet.wait_for(timeout=2.0)
    @patch('lacore.api.get_client_info')
    @defer.inlineCallbacks
    def test_request_simple(self, gci):
        ua = self.getUniqueString()
        gci.return_value = ua
        session = self._makeit()
        for m in ['get', 'post', 'patch']:
            with patch('treq.' + m) as method:
                with patch('treq.content') as content:
                    content.return_value = "{}"
                    rsp = Mock(code=200)
                    url = self.getUniqueString()
                    method.return_value = defer.succeed(rsp)
                    r = yield getattr(session, m)(url)
                    self.assertEqual(r, {})
                    self.assertEqual((url,), method.call_args[0])
                    kwargs = method.call_args[1]

                    self.assertTrue('headers' in kwargs)
                    self.assertTrue('User-Agent' in kwargs['headers'])
                    self.assertEqual(ua, kwargs['headers']['User-Agent'][0])

    @crochet.wait_for(timeout=2.0)
    @defer.inlineCallbacks
    def test_request_simple_user_agent(self):
        ua = self.getUniqueString()
        session = self._makeit(prefs={'user_agent': ua})
        with patch.multiple('treq', get=DEFAULT, content=DEFAULT) as d:
            d['content'].return_value = "{}"
            rsp = Mock(code=200)
            d['get'].return_value = defer.succeed(rsp)
            yield session.get(self.getUniqueString())
            kwargs = d['get'].call_args[1]
            self.assertTrue('headers' in kwargs)
            self.assertTrue('User-Agent' in kwargs['headers'])
            self.assertEqual(ua, kwargs['headers']['User-Agent'][0])

    @crochet.wait_for(timeout=2.0)
    @defer.inlineCallbacks
    def test_request_get_fail(self):
        session = self._makeit()
        with patch('treq.get') as get:
            with patch('treq.content') as content:
                content.return_value = "{}"
                phrase = self.getUniqueString()
                rsp = Mock(code=400, phrase=phrase)
                get.return_value = defer.succeed(rsp)
                url = self.getUniqueString()
                e = yield assert_fails_with(
                    session.get(url), ApiErrorException)
                self.assertEqual(ApiErrorException.msg, str(e))
                self.assertEqual('400 ' + phrase, e.exc)
                self.assertEqual((url,), get.call_args[0])

    @crochet.wait_for(timeout=2.0)
    @defer.inlineCallbacks
    def test_request_get_auth(self):
        auth = (self.getUniqueString(), self.getUniqueString())
        session = self._makeit(
            {'user': auth[0], 'pass': auth[1], 'verify': False})
        with patch('treq.get') as get:
            with patch('treq.content') as content:
                content.return_value = "{}"
                get.side_effect = [
                    defer.succeed(Mock(code=401)),
                    defer.succeed(Mock(code=200))]
                url = self.getUniqueString()
                r = yield session.get(url)
                self.assertEqual(r, {})
                self.assertIn('auth', get.call_args[1])
                self.assertEqual(auth, get.call_args[1]['auth'])

    @crochet.wait_for(timeout=2.0)
    @defer.inlineCallbacks
    def test_request_get_auth_fail(self):
        session = self._makeit()
        with patch('treq.get') as get:
            with patch('treq.content') as content:
                content.return_value = "{}"
                get.side_effect = [
                    defer.succeed(Mock(code=401)),
                    defer.succeed(Mock(code=401))]
                url = self.getUniqueString()
                e = yield assert_fails_with(
                    session.get(url), ApiAuthException)
                self.assertEqual(ApiAuthException.msg, str(e))
                self.assertEqual((url,), get.call_args[0])

    @crochet.wait_for(timeout=2.0)
    @defer.inlineCallbacks
    def test_request_get_404(self):
        session = self._makeit()
        with patch('treq.get') as get:
            with patch('treq.content') as content:
                content.return_value = "{}"
                rsp = Mock(code=404)
                get.return_value = defer.succeed(rsp)
                url = self.getUniqueString()
                e = yield assert_fails_with(
                    session.get(url),
                    ApiUnavailableException)
                self.assertEqual(ApiUnavailableException.msg, str(e))
                self.assertEqual((url,), get.call_args[0])

    @crochet.wait_for(timeout=2.0)
    def test_finalize(self):
        _gets = map(defer.succeed,
                    map(json.loads,
                        [LA_ENDPOINTS_RESPONSE, LA_CAPSULES_RESPONSE]))
        r = self.getUniqueString()
        _patches = imap(defer.succeed, [r])
        sargs = {'get.side_effect': _gets,
                 'patch.side_effect': _patches}
        s = Mock(**sargs)
        from lacore.api import Api
        api = Api(makeprefs()['api'], session=s)
        state = Mock(uri='foo', sandbox=False,
                     capsule={'resource_uri': '/1'})
        op = api.upload(Mock(title='',
                             description='',
                             meta=Mock(size=1)), state)
        d = op.finalize()
        d.addCallback(lambda x: self.assertEqual(r, x))
        d.addCallback(lambda x: self.assertEqual(
            s.patch.call_args[0], ('foo',)))
        d.addCallback(lambda x: self.assertEqual(
            json.loads(s.patch.call_args[1]['data']),
            {
                'status': 'uploaded',
                'parts': 0,
                'size': 1
            }))
        return d



from .base import CipherBase
from Crypto.Cipher import AES
from Crypto.Util import Counter


class CipherAES(CipherBase):

    mode = 'aes-256-ctr'
    BLOCKSIZE = 16

    def __init__(self, key, input=None):
        if not key or len(key) != self.BLOCKSIZE * 2:
            raise ValueError("Invalid key")

        prefix = ''
        nbits = 128
        iv = 0L
        if input:
            prefix = input[:8]    # NIST 800-38A
            iv = long(input[8:].encode('hex'), 16)
            nbits = 64

        self.ctr = Counter.new(nbits, prefix=prefix, initial_value=iv)
        self.obj = AES.new(key, AES.MODE_CTR, counter=self.ctr)

    def encipher_block(self, block):
        # TODO we should let pycrypto do the block handling, it would be way
        # faster
        return self.obj.encrypt(block)

    def decipher_block(self, block):
        # TODO we should let pycrypto do the block handling, it would be way
        # faster
        return self.obj.decrypt(block)

    def flush(self):
        last = super(CipherAES, self).flush()
        return self.obj.encrypt(last)



import os
from . import Dumper
from tempfile import NamedTemporaryFile
from urllib import pathname2url
from lacore.adf.elements import Links


class FileDumper(Dumper):
    tmpargs = {'delete': False,
               'suffix': ".longaccess"}

    def __init__(self, tmpdir='/tmp', home=None, **kwargs):
        super(FileDumper, self).__init__(**kwargs)
        self.tmpargs['dir'] = tmpdir
        self.home = home

    def update(self, result):
        super(FileDumper, self).update(result)
        path = self.path
        if self.home is not None:
            path = os.path.relpath(path, self.home)
        self.docs['links'] = Links(
            local=pathname2url(path))

    def write(self, data):
        self.dst.write(data)

    def __enter__(self):
        self.dst = NamedTemporaryFile(**self.tmpargs)
        self.path = self.dst.name
        return self

    def __exit__(self, eType, eValue, eTrace):
        try:
            super(FileDumper, self).__exit__(eType, eValue, eTrace)
        finally:
            self.dst.close()
            if eType is not None and os.path.exists(self.path):
                os.unlink(self.path)  # don't leave trash



from testtools import TestCase
from mock import patch, Mock
from . import makeprefs


class ApiHttpTest(TestCase):
    def setUp(self):
        super(ApiHttpTest, self).setUp()
        self.prefs = makeprefs()['api']

    def _makeit(self):
        from lacore.api import RequestsFactory
        return RequestsFactory

    @patch('lacore.api.Api')
    def test_request_factory_no_prefs(self, api):
        factory = self._makeit()
        s = self.getUniqueString()
        self.patch(factory, 'TwistedRequestsSession',
                   Mock(return_value=s))
        factory()
        api.assert_called_with({}, s)

    @patch('lacore.api.Api')
    def test_request_factory(self, api):
        factory = self._makeit()
        s = self.getUniqueString()
        self.patch(factory, 'TwistedRequestsSession',
                   Mock(return_value=s))
        p = {self.getUniqueString(): self.getUniqueString()}
        factory(prefs=p)
        api.assert_called_with(p, s)

    @patch('lacore.api.Api')
    def test_request_factory_verify(self, api):
        factory = self._makeit()
        s = self.getUniqueString()
        rs = Mock(return_value=s)
        self.patch(factory, 'TwistedRequestsSession', rs)
        p = {'verify': self.getUniqueString()}
        factory(prefs=p)
        self.assertEqual(1, len(rs.call_args[0]))
        sobj = rs.call_args[0][0]
        self.assertTrue(hasattr(sobj, 'verify'))
        self.assertEqual(sobj.verify, p['verify'])

    @patch('lacore.api.Api')
    def test_request_factory_auth(self, api):
        factory = self._makeit()
        s = self.getUniqueString()
        rs = Mock(return_value=s)
        self.patch(factory, 'TwistedRequestsSession', rs)
        p = {
            'user': self.getUniqueString(),
            'pass': self.getUniqueString()
        }
        factory(prefs=p)
        self.assertEqual(1, len(rs.call_args[0]))
        sobj = rs.call_args[0][0]
        self.assertTrue(hasattr(sobj, 'auth'))
        self.assertEqual(sobj.auth, (p['user'], p['pass']))

    @patch('lacore.api.Api')
    def test_request_factory_useragent(self, api):
        factory = self._makeit()
        s = self.getUniqueString()
        rs = Mock(return_value=s)
        self.patch(factory, 'TwistedRequestsSession', rs)
        p = {'user_agent': self.getUniqueString()}
        factory(prefs=p)
        self.assertEqual(1, len(rs.call_args[0]))
        sobj = rs.call_args[0][0]
        self.assertTrue(hasattr(sobj, 'user_agent'))
        self.assertEqual(sobj.user_agent, p['user_agent'])

    @patch('lacore.api.Api')
    def test_request_factory_session(self, api):
        factory = self._makeit()
        factory()
        args, _ = api.call_args
        self.assertEqual(len(args), 2)
        self.assertEqual(args[0], {})
        self.assertTrue(hasattr(args[1], 'get'))
        self.assertTrue(hasattr(args[1], 'post'))
        self.assertTrue(hasattr(args[1], 'patch'))



from testtools import TestCase
from moto import mock_s3
from boto import connect_s3
from datetime import datetime
from dateutil.relativedelta import relativedelta
from mock import patch


class MPConnectionTest(TestCase):
    @classmethod
    def setup_class(cls):
        cls._token = {
            'token_access_key': '',
            'token_secret_key': '',
            'token_session': '',
            'token_expiration': '',
            'bucket': 'lastage',
            'prefix': 'upload/14/',
        }
        cls._bucket = 'lastage'

    def setUp(self):
        self.s3 = mock_s3()
        self.s3.start()
        boto = connect_s3()
        boto.create_bucket(self._bucket)

        super(MPConnectionTest, self).setUp()

    def tearDown(self):
        self.s3.stop()
        super(MPConnectionTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacore.storage.s3 import MPConnection
        return MPConnection(*args,  **kw)

    def test_mpconnection(self):
        assert self._makeit(**self._token)

    @patch('lacore.storage.s3.getLogger', create=True)
    def test_timeout(self, log):
        tok = self._token
        tok['token_expiration'] = "this is not a timestamp"
        conn = self._makeit(grace=0, **tok)
        self.assertEqual(None, conn.timeout())
        log.assert_called_with()
        log.return_value.debug.assert_called_with(
            'invalid token expiration: %s', 'this is not a timestamp')
        tok['token_expiration'] = datetime.utcnow().isoformat()
        conn = self._makeit(grace=0, **tok)
        self.assertEqual(0, int(conn.timeout()))
        tok['token_expiration'] = datetime.utcnow()
        conn = self._makeit(grace=0, **tok)
        self.assertEqual(0, int(conn.timeout()))
        tok['token_expiration'] = datetime.utcnow()+relativedelta(seconds=100)
        conn = self._makeit(grace=100, **tok)
        self.assertEqual(0, int(conn.timeout()))
        tok['token_expiration'] = datetime.utcnow()+relativedelta(seconds=101)
        conn = self._makeit(grace=0, **tok)
        self.assertEqual(100, int(conn.timeout()))



from testtools import TestCase
from mock import patch, DEFAULT


class MiscTest(TestCase):
    def test_get_client_info(self):
        from lacore import get_client_info, __version__
        self.assertEqual('lacore ' + __version__, get_client_info(True, False))
        self.assertEqual("Longaccess Python SDK " + __version__,
                         get_client_info(False, False))
        patches = ['platform', 'python_version', 'machine']
        kw = dict([(p, DEFAULT) for p in patches])
        with patch.multiple('platform', **kw) as mocks:
            expected = ["lacore", __version__]
            for p in patches:
                v = self.getUniqueString()
                expected.append("py" + v if p is "python_version" else v)
                mocks[p].return_value = v
            self.assertEqual(" ".join(expected), get_client_info(True, True))
            mocks['platform'].assert_called_with(True, True)
        with patch.multiple('platform', **kw) as mocks:
            expected = ["Longaccess Python SDK", __version__]
            for p in patches:
                v = self.getUniqueString()
                expected.append("py" + v if p is "python_version" else v)
                mocks[p].return_value = v
            self.assertEqual(" ".join(expected), get_client_info(False, True))
            mocks['platform'].assert_called_with(True, False)



from twisted.python.failure import Failure
from twisted.internet import reactor, defer
from twisted.internet.defer import setDebugging as debugTwisted
from twisted.python import log as twisted_log
from crochet import setup, run_in_reactor, TimeoutError
from functools import update_wrapper, wraps, partial
from contextlib import contextmanager


def block_timeout(timeout):
    def _block(f):
        """ Decorate a method to block in crochet reactor
        """
        fblocking = run_in_reactor(f)

        @wraps(f)
        def wrap(*args, **kwargs):
            if not reactor.running:
                setup()
            result = fblocking(*args, **kwargs)
            try:
                return result.wait(timeout=timeout)
            except TimeoutError:
                result.cancel()
                raise
        return wrap
    return _block


def block(f):
    return block_timeout(None)(f)


class deferred_property(object):
    def __init__(self, f):
        self.f = defer.inlineCallbacks(f)
        update_wrapper(self, self.f)

    def update(self, obj, name, value):
        if isinstance(value, Failure):
            del obj.__dict__[name]
        else:
            obj.__dict__[name] = value
        return value

    def __get__(self, obj, cls):
        if obj is None:
            return self
        name = self.f.__name__
        if name not in obj.__dict__:
            obj.__dict__[name] = self.f(obj)
            obj.__dict__[name].addBoth(
                partial(self.update, obj, name))
        return obj.__dict__[name]

    def __set__(self, obj, value):
        obj.__dict__[self.f.__name__] = value


@contextmanager
def twisted_log_observer(level):
    if level > 2:
        debugTwisted(True)
    observer = twisted_log.PythonLoggingObserver()
    observer.start()

    yield None

    observer.stop()



# -*- coding: utf-8 -*-

from testtools import TestCase


class DateTest(TestCase):
    def setUp(self):
        super(DateTest, self).setUp()

    def tearDown(self):
        super(DateTest, self).tearDown()

    def test_epoch(self):
        from lacore.date import epoch

        self.assertEqual("1970-01-01T00:00:00+00:00", epoch().isoformat())



import os
import hashlib

from .. import StorageConnection
from lacore.date import parse_timestamp, remaining_time
from lacore.log import getLogger
from lacore.decorators import cached_property
from lacore.exceptions import UploadInvalidError
from boto import connect_s3
from boto.s3.key import Key
from boto.s3.connection import NoHostProvided
from boto.regioninfo import get_regions


class S3Connection(StorageConnection):
    def __init__(self, token_access_key=None, token_secret_key=None,
                 token_session=None, token_expiration=None, bucket=None,
                 prefix='', port=None, host=NoHostProvided,
                 **kwargs):
        super(S3Connection, self).__init__(**kwargs)
        self.accesskey = token_access_key
        self.secret = token_secret_key
        self.sectoken = token_session
        self.expires = token_expiration
        self._bucket = bucket
        self.prefix = prefix
        self.conn = None
        self.port = port
        if self.port is None:
            port = os.environ.get('S3_PORT')
            self.port = int(port) if port is not None else None
        self.host = host
        if self.host is NoHostProvided:
            host = os.environ.get('S3_HOST')
            s3_region = os.environ.get('S3_REGION')
            if host is None and s3_region is not None:
                for region in get_regions('s3'):
                    if region.name == s3_region:
                        self.host = region.endpoint

    def getconnection(self):
        if self.conn is None:
            self.conn = connect_s3(
                aws_access_key_id=self.accesskey,
                aws_secret_access_key=self.secret,
                security_token=self.sectoken,
                host=self.host,
                port=self.port)
        return self.conn

    @cached_property
    def bucket(self):
        return self.getconnection().get_bucket(self._bucket)

    def newkey(self, key):
        return Key(self.bucket, self.prefix + key)

    def __getstate__(self):
        """ prepare the object for pickling. """
        state = self.__dict__
        state['conn'] = None
        return state

    def newupload(self, key):
        return self.bucket.initiate_multipart_upload(self.prefix + key)

    def getupload(self, uid):
        for upload in self.bucket.get_all_multipart_uploads():
            if uid == upload.id:
                return upload

    def complete_multipart(self, upload, etags):
        xml = '<CompleteMultipartUpload>\n'
        md5 = hashlib.md5()
        for seq, etag in enumerate(etags):
            xml += '  <Part>\n'
            xml += '    <PartNumber>{}</PartNumber>\n'.format(seq+1)
            xml += '    <ETag>{}</ETag>\n'.format(etag)
            xml += '  </Part>\n'
            md5.update(etag.replace('"', '').decode("hex"))
        xml += '</CompleteMultipartUpload>'
        key = self.bucket.complete_multipart_upload(
            upload.key_name, upload.id, xml)
        if key.etag != '"{0}-{1}"'.format(md5.hexdigest(), len(etags)):
            raise UploadInvalidError("the multipart key had an invalid eTag")
        return key


class MPConnection(S3Connection):
    def __init__(self, token_expiration=None, grace=120, **kwargs):
        super(MPConnection, self).__init__(**kwargs)
        self.grace = grace
        self.conn = None
        self.expiration = None
        if token_expiration is not None:
            try:
                self.expiration = parse_timestamp(token_expiration)
            except ValueError:
                getLogger().debug("invalid token expiration: %s",
                                  token_expiration)

    def timeout(self):
        """ return total number of seconds till
            this connection must be renewed. """
        if not self.expiration:
            return None
        remaining = remaining_time(self.expiration)
        return remaining.total_seconds() - self.grace



from testtools import TestCase, ExpectedException
from mock import Mock, patch
from itertools import repeat
from . import makeprefs
from lacore.exceptions import (ApiAuthException, ApiErrorException,
                               UploadEmptyError)
from lacore.async import block
from twisted.internet import defer
from itertools import imap
import json


class ApiTest(TestCase):
    def setUp(self):
        super(ApiTest, self).setUp()
        self.prefs = makeprefs()['api']

    def tearDown(self):
        super(ApiTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacore.api import Api
        return Api(*args, **kw)

    def _mocksessions(self, rsps):
        return Mock(**rsps)

    def _json_request(self, url, mock_call):
        args, kwargs = mock_call.call_args
        headers = kwargs['headers']
        self.assertEqual('application/json', headers['content-type'])
        self.assertEqual(url, args[0])
        return kwargs['data']

    def test_api(self):
        assert self._makeit(self.prefs, Mock())

    def test_api_root(self):
        r = json.loads(LA_ENDPOINTS_RESPONSE)
        s = self._mocksessions({'get.return_value': defer.succeed(r)})
        api = self._makeit(self.prefs, session=s)
        self.assertEqual(r, api.root)

    def test_no_capsules(self):
        caps = json.loads(LA_EMPTY_COLLECTION)
        r = [defer.succeed(json.loads(LA_ENDPOINTS_RESPONSE)),
             defer.succeed(caps)]
        s = self._mocksessions({'get.side_effect': r})
        api = self._makeit(self.prefs, session=s)
        capsules = list(api.capsules())
        self.assertEqual(len(capsules), 0)

    def simple_session(self, response):
        j = map(defer.succeed,
                map(json.loads, [LA_ENDPOINTS_RESPONSE, LA_CAPSULES_RESPONSE]))
        return self._mocksessions({'get.side_effect': j})

    def capsule_session(self):
        return self.simple_session(LA_CAPSULES_RESPONSE)

    def test_capsules(self):
        api = self._makeit(self.prefs, session=self.capsule_session())
        capsules = list(api.capsules())
        self.assertEqual(len(capsules), 2)

    def test_capsule_ids(self):
        api = self._makeit(self.prefs, session=self.capsule_session())
        ids = api.capsule_ids()
        self.assertTrue(2 in ids)
        self.assertTrue(3 in ids)

    def test_capsule_ids_size(self):
        api = self._makeit(self.prefs, session=self.capsule_session())
        ids = api.capsule_ids(150).iterkeys()
        self.assertTrue(2 in ids)
        self.assertTrue(3 not in ids)

    def test_no_archives(self):
        arcs = json.loads(LA_EMPTY_COLLECTION)
        r = [defer.succeed(json.loads(LA_ENDPOINTS_RESPONSE)),
             defer.succeed(arcs)]
        s = self._mocksessions({'get.side_effect': r})
        api = self._makeit(self.prefs, session=s)
        archives = list(api.archives())
        self.assertEqual(len(archives), 0)

    def archive_session(self):
        return self.simple_session(LA_ARCHIVES_RESPONSE)

    def test_archives(self):
        api = self._makeit(self.prefs, session=self.archive_session())
        archives = list(api.archives())
        self.assertEqual(len(archives), 2)

    def test_unauthorized(self):
        r = [defer.succeed(json.loads(LA_ENDPOINTS_RESPONSE)),
             defer.fail(ApiAuthException())]
        s = self._mocksessions({'get.side_effect': r})
        api = self._makeit(self.prefs, session=s)
        with ExpectedException(ApiAuthException):
            list(api.capsules())

    def test_status_start(self):
        _gets = map(defer.succeed,
                    map(json.loads,
                        [LA_ENDPOINTS_RESPONSE, LA_CAPSULES_RESPONSE]))
        _posts = imap(defer.succeed, repeat(json.loads(LA_UPLOAD_RESPONSE)))
        s = self._mocksessions({'get.side_effect': _gets,
                                'post.side_effect': _posts})
        api = self._makeit(self.prefs, session=s)
        state = Mock(uri=None, sandbox=False,
                     capsule={'resource_uri': '/1'})
        op = api.upload(Mock(title='',
                             description='',
                             meta=Mock(size=1)), state)
        self.assertEqual(None, op.uri)
        block(lambda: op.status)()
        uri = 'http://baz.com/api/v1/upload/1/'
        self.assertEqual(uri, op.uri)
        self.assertFalse(op.sandbox)

    def test_status_start_sandbox(self):
        _gets = map(defer.succeed,
                    map(json.loads,
                        [LA_ENDPOINTS_RESPONSE, LA_CAPSULES_RESPONSE]))
        _posts = imap(defer.succeed, repeat(json.loads(LA_SANDBOX_RESPONSE)))
        s = self._mocksessions({'get.side_effect': _gets,
                                'post.side_effect': _posts})
        api = self._makeit(self.prefs, session=s)
        state = Mock(uri=None, sandbox=True,
                     capsule={'resource_uri': '/1'})
        op = api.upload(Mock(title='',
                             description='',
                             meta=Mock(size=1)), state)
        self.assertEqual(None, op.uri)
        block(lambda: op.status)()
        uri = 'http://baz.com/api/v1/upload/1/'
        self.assertEqual(uri, op.uri)
        self.assertTrue(op.sandbox)

    @patch('lacore.api.getLogger', create=True)
    def test_status_error(self, log):
        _gets = map(defer.succeed,
                    map(json.loads,
                        [LA_ENDPOINTS_RESPONSE, LA_CAPSULES_RESPONSE,
                         LA_UPLOAD_RESPONSE]))
        _gets.append(defer.fail(ApiErrorException()))
        _posts = imap(defer.succeed, repeat(json.loads(LA_UPLOAD_RESPONSE)))
        s = self._mocksessions({'get.side_effect': _gets,
                                'post.side_effect': _posts})
        api = self._makeit(self.prefs, session=s)
        state = Mock(uri=None, sandbox=False,
                     capsule={'resource_uri': '/1'})
        op = api.upload(Mock(title='',
                             description='',
                             meta=Mock(size=1)), state)
        block(lambda: op.status)()
        block(lambda: op.status)()
        block(lambda: op.status)()
        log.return_value.debug.return_value = None
        self.assertRaises(ApiErrorException, block(lambda: op.status))
        log.assert_called_with()
        log.return_value.debug.assert_called_with(
            'Exception while getting status', exc_info=True)

    def test_finalize_empty(self):
        api = self._makeit(self.prefs)
        op = api.upload(Mock(), Mock(uri=None, capsule={'resource_uri': '/1'}))
        self.assertRaises(UploadEmptyError, op.finalize)


LA_UPLOAD_RESPONSE_TMPL = """{{
    "id": 1,
    "capsule": "/api/v1/capsule/1/",
    "title": "foo",
    "description": "foo",
    "resource_uri": "/api/v1/upload/1/",
    "status": "pending",
    "token_access_key": "XXXXXXXXXXXXXXXXXXXX",
    "token_secret_key": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
    "token_session": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
    "token_expiration": "",
    "token_uid": "arn:aws:iam::XXXXXXXXXXXX:user/sts-test-dummy",
    "bucket": "lastage",
    "prefix": "prefix",
    "sandbox": {sandbox}
}}
"""


LA_UPLOAD_RESPONSE = LA_UPLOAD_RESPONSE_TMPL.format(sandbox='false')


LA_SANDBOX_RESPONSE = LA_UPLOAD_RESPONSE_TMPL.format(sandbox='true')


LA_ENDPOINTS_RESPONSE = """{
  "account": {
     "list_endpoint": "/api/v1/account/",
     "schema": "/api/v1/account/schema/"
  },
  "archive": {
     "list_endpoint": "/api/v1/archive/",
     "schema": "/api/v1/archive/schema/"
  },
  "capsule": {
     "list_endpoint": "/api/v1/capsule/",
     "schema": "/api/v1/capsule/schema/"
  },
  "upload": {
     "list_endpoint": "/api/v1/upload/",
     "schema": "/api/v1/upload/schema/"
  }
}
"""

LA_CAPSULES_RESPONSE = """{
  "meta": {
     "limit": 20,
     "next": null,
     "offset": 0,
     "previous": null,
     "total_count": 2
  },
  "objects": [
     {
         "created": "2013-06-07T10:45:01",
         "id": 3,
         "resource_uri": "/api/v1/capsule/3/",
         "title": "Photos",
         "size": 1000,
         "remaining": 100,
         "user": "/api/v1/user/3/"
      },
      {
          "created": "2013-06-07T10:44:38",
          "id": 2,
          "resource_uri": "/api/v1/capsule/2/",
          "title": "Stuff",
          "size": 1000,
          "remaining": 200,
          "user": "/api/v1/user/2/"
       }
    ]
}"""

LA_EMPTY_COLLECTION = """{
  "meta": {
     "limit": 20,
     "next": null,
     "offset": 0,
     "previous": null,
     "total_count": 0
  },
  "objects": []
}"""

LA_ARCHIVES_RESPONSE = """{
   "objects" : [
      {
         "capsule" : "/api/v1/capsule/3/",
         "description" : "man",
         "size" : "1310288",
         "key" : "120-NV5DVV",
         "created" : "2014-04-10T06:19Z",
         "resource_uri" : "/api/v1/archive/120-NV5DVV/",
         "title" : "yo",
         "expires" : "2015-01-21T23:59Z"
      },
      {
         "capsule" : "/api/v1/capsule/2/",
         "description" : "baba",
         "size" : "3150096",
         "key" : "109-PW9SLN",
         "created" : "2014-04-05T16:04Z",
         "resource_uri" : "/api/v1/archive/109-PW9SLN/",
         "title" : "yoyo",
         "expires" : "2015-01-21T23:59Z"
      }
   ],
   "meta" : {
      "previous" : null,
      "next" : "/api/v1/archive/?limit=20&offset=20",
      "limit" : 20,
      "total_count" : 2,
      "offset" : 0
   }
}"""



from Crypto import Random
from ..decorators import contains
from .base import CipherBase
from . import aes  # NOQA
from . import xor  # NOQA
from . import none  # NOQA


@contains(dict)
def cipher_modes():
    for c in CipherBase.__subclasses__():
        yield (c.mode, c)


def new_key(nbit):
    return Random.new().read(32)


def get_cipher(archive, cert):
    """ return a suitable cipher object to decrypt/encrypt the archive
        using the certificate key and input parameters (iv, etc).
    """
    input = None
    key = None
    cipher = archive.meta.cipher
    if hasattr(cipher, 'input'):
        input = cipher.input
    if hasattr(cert, 'key'):
        key = cert.key
    elif hasattr(cert, 'keys') and len(cert.keys) > 0:
        if hasattr(cipher, 'key'):
            idx = cipher.key
            if idx > 0 and idx <= len(cert.keys):
                key = cert.keys[idx-1]
        else:
            key = cert.keys[0]
    mode = None
    if hasattr(cipher, 'mode'):
        mode = cipher.mode
    elif isinstance(cipher, str):
        mode = cipher
    if mode:
        cipher_class = cipher_modes().get(mode)
        if cipher_class:
            return cipher_class(key, input)



import logging
from boto import config as boto_config
from contextlib import contextmanager


simplefmt = '%(name)-15s %(levelname)-8s %(processName)-10s %(message)s'
logging.basicConfig(format=simplefmt)


def setupLogging(level=0, logfile=None):
    # this overrides any user debug setting in the boto configuration
    if not boto_config.has_section('Boto'):
        boto_config.add_section('Boto')

    import httplib
    if level > 4:
        httplib.HTTPConnection.debuglevel = 1

    if level == 0:
        level = 'WARN'
        boto_config.set('Boto', 'debug', '0')
    elif level == 1:
        level = 'INFO'
        boto_config.set('Boto', 'debug', '0')
    elif level == 2:
        level = 'DEBUG'
        boto_config.set('Boto', 'debug', '1')
    else:
        level = 'DEBUG'
        boto_config.set('Boto', 'debug', '2')

    logconf = {
        'version': 1,
        'formatters': {
            'simple': {
                'class': 'logging.Formatter',
                'format': simplefmt,
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': level,
                'formatter': 'simple',
            }
        },
        'loggers': {
            'boto': {
                'handlers': ['console'],
                'propagate': True
            },
            'lacore': {
                'handlers': ['console'],
                'propagate': True
            },
            'multiprocessing': {
                'handlers': ['console'],
                'propagate': True
            },
            'requests.packages.urllib3': {
                'handlers': ['console'],
                'propagate': True
            },
            'twisted': {
                'handlers': ['console'],
                'propagate': True
            }
        },
        'root': {'level': level}
    }

    logging.config.dictConfig(logconf)

    return log_open(logfile)


@contextmanager
def log_open(log):
    logfile = None
    if log is not None:
        try:
            logfile = open(log, 'w+')
            handler = logging.StreamHandler(logfile)
            handler.setFormatter(logging.Formatter(simplefmt))
            logging.getLogger('').addHandler(handler)
        except Exception:
            getLogger().debug("couldn't open log", exc_info=True)

    yield None

    if logfile is not None:
        logfile.close()


def getLogger(logger='lacore'):
    return logging.getLogger(logger)



from abc import ABCMeta, abstractmethod


class StorageConnection(object):
    __metaclass__ = ABCMeta

    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def newkey(self, key):
        """ return an object representing a storage object """



import io
import requests
import os.path

from itertools import imap
from .stream import StreamArchiver
from contextlib import closing


def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE):
    """
    Lets you use an iterable (e.g. a generator) that yields bytestrings as a
    read-only input stream.

    The stream implements Python 3's newer I/O API (available in Python 2's io
    module).  For efficiency, the stream is buffered.

    From: http://stackoverflow.com/a/20260030/729491
    """
    class IterStream(io.RawIOBase):
        def __init__(self):
            self.leftover = None

        def readable(self):
            return True

        def readinto(self, b):
            try:
                l = len(b)  # We're supposed to return at most this much
                chunk = self.leftover or next(iterable)
                output, self.leftover = chunk[:l], chunk[l:]
                b[:len(output)] = output
                return len(output)
            except StopIteration:
                return 0    # indicate EOF
    return io.BufferedReader(IterStream(), buffer_size=buffer_size)


class UrlArchiver(StreamArchiver):
    url_chunk = io.DEFAULT_BUFFER_SIZE

    def __init__(self, url_chunk=None, progress_cb=None, *args, **kwargs):
        if url_chunk is not None:
            self.url_chunk = url_chunk
        self.progress_cb = progress_cb
        super(UrlArchiver, self).__init__(*args, **kwargs)

    def closing_url(self, url, rel):
        with closing(requests.get(url, stream=True)) as r:
            for c in r.iter_content(chunk_size=self.url_chunk):
                yield c

    def url_to_stream(self, item):
        try:
            url, rel = item
        except:
            url = item
            rel = os.path.basename(item.strip('/'))
        r = iterable_to_stream(self.closing_url(url, rel))
        return (r, rel)

    def args(self, items, cb):
        return super(UrlArchiver, self).args(
            imap(self.url_to_stream, items), cb)



import os
import httpretty

from . import makeprefs
from testtools import TestCase, ExpectedException
from unittest import SkipTest
from mock import Mock
from boto import connect_s3
from moto import mock_s3


class S3DumperTest(TestCase):
    @classmethod
    def setup_class(cls):
        cls._bucket = 'lastage2'

    def setUp(self):
        super(S3DumperTest, self).setUp()
        self.prefs = makeprefs()
        self.home = os.path.join('t', 'data', 'home')
        self.s3 = mock_s3()
        self.s3.start()
        boto = connect_s3()
        boto.create_bucket(self._bucket)

    def tearDown(self):
        self.s3.stop()
        super(S3DumperTest, self).tearDown()

    def test_s3_constructor(self):
        from lacore.dumper.s3 import S3Dumper
        from lacore.storage.s3 import S3Connection
        dest = S3Dumper()
        self.assertTrue(isinstance(dest.connection, S3Connection))

    def test_s3_constructor_min_chunk(self):
        from lacore.dumper.s3 import S3Dumper

        msg = "S3Dumper chunk must be at least 5242880"
        with ExpectedException(ValueError, msg):
            S3Dumper(chunk=123)

    def test_s3_constructor_with_connection(self):
        from lacore.dumper.s3 import S3Dumper
        mockc = Mock()
        dest = S3Dumper(connection=mockc)
        self.assertEqual(dest.connection, mockc)

    def test_s3_dumper(self):
        try:
            import zipstream  # noqa
        except ImportError:
            raise SkipTest("requires python-zipstream")
        httpretty.register_uri(httpretty.GET, "http://foobar/",
                               body='file contents man',
                               content_type="text/plain")
        from lacore.dumper.s3 import S3Dumper
        from lacore.storage.s3 import S3Connection
        c = S3Connection(bucket=self._bucket)
        dest = S3Dumper(key='baz', connection=c, title='foo')
        cb = Mock()
        list(dest.dump([('http://foobar', 'foobar')], cb))
        self.assertEqual(160, dest.docs['archive'].meta.size)
        from hashlib import md5
        m = md5()
        m.update(c.bucket.lookup('baz').get_contents_as_string())
        self.assertEqual(m.digest().encode('hex'),
                         dest.docs['auth'].md5.encode('hex'))

    def test_s3_stream_dumper(self):
        try:
            import zipstream  # noqa
        except ImportError:
            raise SkipTest("requires python-zipstream")
        httpretty.register_uri(httpretty.GET, "http://foobar/",
                               body='file contents man',
                               content_type="text/plain")
        from lacore.dumper.s3 import S3StreamDumper
        from lacore.storage.s3 import S3Connection
        from urllib2 import urlopen
        c = S3Connection(bucket=self._bucket)
        dest = S3StreamDumper(key='baz', connection=c, title='foo')
        cb = Mock()
        list(dest.dump([(urlopen('http://foobar'), 'foobar')], cb))
        self.assertEqual(160, dest.docs['archive'].meta.size)
        from hashlib import md5
        m = md5()
        m.update(c.bucket.lookup('baz').get_contents_as_string())
        self.assertEqual(m.digest().encode('hex'),
                         dest.docs['auth'].md5.encode('hex'))



import hashlib
import io

from lacore.adf.elements import Auth


class HashingMixin(object):
    def __init__(self, hashf='md5', *args, **kwargs):
        self.hashf = hashf
        self.hashobj = getattr(hashlib, hashf)()
        super(HashingMixin, self).__init__(*args, **kwargs)

    def update(self, data):
        self.hashobj.update(data)

    def digest(self):
        return self.hashobj.digest()

    @property
    def hashobj(self):
        return getattr(self, "_hash_" + self.hashf)

    @hashobj.setter
    def hashobj(self, hobj):
        setattr(self, "_hash_" + self.hashf, hobj)


class HashFile(HashingMixin, io.FileIO):
    def read(self, n):
        d = super(HashFile, self).read(n)
        if d is not None:
            self.hashobj.update(d)
        return d


class MyHashObj(HashingMixin):

    def __init__(self, hashf='sha512', *args, **kwargs):
        super(MyHashObj, self).__init__(hashf=hashf, *args, **kwargs)
        self.md5 = hashlib.md5()

    def update(self, data):
        self.md5.update(data)
        super(MyHashObj, self).update(data)

    def auth(self):
        args = {'md5': self.md5.digest()}
        args[self.hashf] = self.digest()
        return Auth(**args)

    def verify(self, auth):
        ret = (auth.md5 == self.md5.digest())
        if ret and hasattr(auth, self.hashf):
            h = getattr(auth, self.hashf)
            ret = (self.digest() == h)
        return ret



import versioneer

versioneer.VCS = 'git'
versioneer.versionfile_source = 'lacore/version.py'
versioneer.versionfile_build = 'lacore/version.py'
versioneer.tag_prefix = ''
versioneer.parentdir_prefix = 'lacore-'
from setuptools import setup, find_packages


try:
    import pandoc
    pandoc.core.PANDOC_PATH = 'pandoc'
    readme = pandoc.Document()
    readme.markdown = open('README.md').read()
    description = readme.rst
except (IOError, ImportError):
    description = 'This is the prototype client program' + \
        'for interacting with the Longaccess service.'


def pep386adapt(version):
    if version is not None and '-' in version:
        # adapt git-describe version to be in line with PEP 386
        parts = version.split('-')
        parts[-2] = 'post'+parts[-2]
        version = '.'.join(parts[:-1])
    return version


setup(version=pep386adapt(versioneer.get_version()),
      name="lacore",
      author="Konstantinos Koukopoulos",
      author_email='kk@longaccess.com',
      description="The Long Access Python SDK - Core",
      long_description=description,
      url='http://github.com/longaccess/longaccess-core/',
      license='Apache',
      packages=find_packages(exclude=['features*', '*.t']),
      install_requires=[
          'boto>=2.14, <3',
          'filechunkio>=1.5',
          'requests>=2.0.0, <3',
          'pycrypto>=2.6, <2.7',
          'PyYAML>=3.10, <4',
          'Twisted>=13.2, <14',
          'crochet>=1.0.0, <2',
          'treq>=0.2.0, <0.3',
          'python-dateutil>=2.2, <3'
          ],
      tests_require=['testtools'],
      test_suite="lacore.t",
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Intended Audience :: Information Technology',
          'Natural Language :: English',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Topic :: System :: Archiving',
          'Topic :: Utilities',
          ],
      cmdclass=versioneer.get_cmdclass(),
      )



import yaml


class PrettySafeLoader(yaml.SafeLoader):
    pass


class BaseYAMLObject(yaml.YAMLObject):
    yaml_loader = PrettySafeLoader
    yaml_dumper = yaml.SafeDumper

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_mapping(cls.yaml_tag, data.__dict__)


def add_path_resolver(tag, keys):
    yaml.add_path_resolver(tag, keys, dict, Loader=PrettySafeLoader)


def load_file(f):
    try:
        return yaml.load_all(f, Loader=PrettySafeLoader,
                             tz_aware_datetimes=True)
    except TypeError:
        return yaml.load_all(f, Loader=PrettySafeLoader)



import pickle

from testtools import TestCase, ExpectedException
from moto import mock_s3
from mock import patch, Mock
import boto


class S3ConnectionTest(TestCase):
    @classmethod
    def setup_class(cls):
        cls._token = {
            'token_access_key': '',
            'token_secret_key': '',
            'token_session': '',
            'token_expiration': '',
            'bucket': 'lastage',
            'prefix': 'upload/14/',
        }

    def setUp(self):
        self.s3 = mock_s3()
        self.s3.start()
        boto.connect_s3().create_bucket(self._token['bucket'])

        super(S3ConnectionTest, self).setUp()

    def tearDown(self):
        self.s3.stop()
        super(S3ConnectionTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacore.storage.s3 import S3Connection
        return S3Connection(*args,  **kw)

    def test_s3connection(self):
        assert self._makeit(**self._token)

    def test_getconnection(self):
        conn = self._makeit(**self._token)
        assert conn.getconnection()

    def test_getbucket(self):
        conn = self._makeit(**self._token)
        self.assertEqual(conn.bucket.name, self._token['bucket'])

    @patch('lacore.storage.s3.connect_s3')
    def test_getbucket_exc(self, mc):
        from boto.exception import S3ResponseError
        mc.return_value.get_bucket.side_effect = S3ResponseError(
            '401', 'FooError', '')
        conn = self._makeit(**self._token)
        with ExpectedException(S3ResponseError,
                               "S3ResponseError: 401 FooError"):
            print conn.bucket

    def test_getbucket_404(self):
        from boto.exception import S3ResponseError
        tok = {}
        tok.update(self._token)
        tok['bucket'] = 'nosuchbucket'
        conn = self._makeit(**tok)
        with ExpectedException(S3ResponseError,
                               "S3ResponseError: 404 Not Found"):
            print conn.bucket

    def test_newkey(self):
        conn = self._makeit(**self._token)
        key = conn.newkey('foobar')
        self.assertEqual(conn.bucket.name, key.bucket.name)
        self.assertEqual('upload/14/foobar', key.name)

    def test_pickle(self):
        conn = self._makeit(**self._token)
        conn.getconnection()
        self.assertTrue(conn.conn is not None)
        conn = pickle.loads(pickle.dumps(conn))
        self.assertTrue(conn.conn is None)

    def test_multipart_upload(self):
        conn = self._makeit(**self._token)
        u = conn.newupload('foo')
        u2 = conn.getupload(u.id)
        self.assertEqual(u.id, u2.id)
        self.assertEqual(u.key_name, u2.key_name)
        self.assertTrue(conn.getupload('lala') is None)

    @patch.object(boto.s3.bucket.Bucket, 'complete_multipart_upload')
    def test_complete_multipart(self, mock_cu):
        conn = self._makeit(**self._token)
        u = conn.newupload('foo')
        mock_cu.return_value = Mock(etag='foo')
        from lacore.exceptions import UploadInvalidError
        with ExpectedException(UploadInvalidError):
            conn.complete_multipart(u, ['abcd'])



from __future__ import division

import os
import math

from sys import maxint
from itertools import imap
from tempfile import mkstemp
from boto.utils import compute_md5
from filechunkio import FileChunkIO


class FileHash(object):
    def __get__(self, instance, owner):
        try:
            return (instance.md5, instance.b64md5)
        except AttributeError:
            instance.seek(0)
            (instance.md5, instance.b64md5, size) = compute_md5(
                instance, size=instance.bytes)
            instance.seek(0)
            return (instance.md5, instance.b64md5)


class SavedPart(file):
    def __init__(self, source, num, *args, **kwargs):
        super(SavedPart, self).__init__(*args, **kwargs)
        self.bytes = source.chunksize(num)

    hash = FileHash()


class FilePart(FileChunkIO):

    def __init__(self, source, num, *args, **kwargs):
        super(FilePart, self).__init__(
            source.path, 'r',
            offset=source.chunkstart(num),
            bytes=source.chunksize(num), *args, **kwargs)

    hash = FileHash()


class ChunkedFile(object):
    minchunk = 5242880
    maxchunk = 104857600

    def __init__(self, path, skip=0, chunk=None):
        self.path = path
        if not os.path.exists(path):
            raise IOError("File {} not found".format(path))
        self.isfile = os.path.isfile(path)
        size = maxint
        if self.isfile:
            size = os.path.getsize(path)
        assert size >= skip
        self.skip = skip
        self.size = size-skip
        self.chunk = chunk
        if self.chunk is None:
            chunk = int(self.size/100)
            maxchunk = max(chunk, self.minchunk)
            self.chunk = min(maxchunk, self.maxchunk)
        self.chunks = int(math.ceil(self.size/self.chunk))
        if self.chunks == 0:
            self.chunks = 1

    def _savedchunks(self, tempdir):
        # split file and save parts to disk
        f = open(self.path, "rb")

        def _save(seq):
            b = f.read(self.chunk)
            if b:
                prefix = "part-{:>4}".format(seq)
                fh, fn = mkstemp(dir=tempdir, prefix=prefix)
                os.write(fh, b)
                os.close(fh)
                return fn
            f.close()

        return imap(_save, xrange(self.chunks))

    def chunkstart(self, num):
        if num >= self.chunks:
            raise ValueError("chunk number out of range")
        return self.skip + num * self.chunk

    def chunksize(self, num):
        start = self.chunkstart(num) - self.skip
        if (start + self.chunk > self.size):
            # chunk end is EOF
            return self.size - start
        else:
            return self.chunk

    def chunkfile(self, seq, fname):
        if self.isfile:
            return FilePart(self, seq)
        else:
            return SavedPart(self, seq, fname)



# -*- coding: utf-8 -*-
from testtools import TestCase, ExpectedException
from nose.tools import raises
from mock import Mock, MagicMock


class CryptIOTest(TestCase):
    def setUp(self):
        super(CryptIOTest, self).setUp()

    def tearDown(self):
        super(CryptIOTest, self).tearDown()

    def _makeit(self, *args, **kwargs):
        from lacore.crypt import CryptIO
        return CryptIO(*args, **kwargs)

    def _mockfile(self, mode='rb', data=[]):
        mf = MagicMock(mode=mode, write=Mock(), read=Mock(side_effect=data))
        mf.__iter__.return_value = data
        return mf

    def _makecipher(self):
        return Mock(encipher=lambda x: x,
                    decipher=lambda x, y=None: x,
                    BLOCKSIZE=32,
                    flush=lambda: '')

    def test_cryptio(self):
        self.assertRaises(ValueError, self._makeit, None, None)
        self.assertRaises(ValueError, self._makeit, None, None, mode='w')
        self.assertRaises(ValueError, self._makeit, None, Mock(), mode='w')
        mockf = self._mockfile(mode='wb')
        self.assertRaises(ValueError, self._makeit, mockf, None)
        obj = self._makeit(mockf, Mock())
        self.assertTrue(obj.writable())
        obj = self._makeit(self._mockfile(), Mock(), mode='rU')
        self.assertTrue(obj.readable())
        dec = Mock()
        del dec.decipher
        mockf = self._mockfile()
        self.assertRaises(ValueError, self._makeit, mockf, dec)
        self.assertRaises(IOError, self._makeit, mockf, dec, mode='x')

    def test_cryptio_nomode(self):
        f = self._mockfile()
        del f.mode
        self._makeit(f, Mock())

    def test_cryptio_badmode(self):
        self.assertRaises(IOError, self._makeit, Mock(mode='foo'), Mock())

    def test_write_to_read(self):
        f = self._makeit(self._mockfile(), Mock())
        self.assertRaises(IOError, f.write, "")

    def test_read_from_write(self):
        f = self._makeit(self._mockfile('wb'), Mock())
        self.assertRaises(IOError, f.read)

    @raises(ValueError)
    def test_cryptio_write_closed(self):
        f = self._mockfile(mode='wb')
        c = self._makeit(f, Mock())
        c.close()
        c.close()
        c.write("FOO")

    @raises(ValueError)
    def test_cryptio_read_closed(self):
        f = self._mockfile()
        c = self._makeit(f, Mock())
        c.close()
        c.read()

    def test_cryptio_memoryview(self):
        f = self._mockfile(mode='wb')
        c = self._makeit(f, Mock())
        c.write(memoryview("FOO"))

    def test_read(self):
        fdata = ['0'*1024 for _ in range(5)]
        f = self._mockfile(data=fdata)
        c = self._makeit(f, self._makecipher())
        out = c.read(500)
        self.assertEqual(500, len(out))
        out += c.read(524)
        self.assertEqual(1024, len(out))
        out += c.read()
        self.assertEqual(5*1024, len(out))
        out += c.read()
        self.assertEqual(5*1024, len(out))

    def test_write(self):
        import StringIO
        f = StringIO.StringIO()
        c = self._makeit(f, self._makecipher(), mode='wb')
        d = '0' * 100
        for _ in range(5):
            c.write(d)
        c.flush()
        self.assertEqual(500, len(f.getvalue()))

    def test_write_context(self):
        import StringIO
        f = StringIO.StringIO()
        ciph = self._makecipher()
        ciph.flush = Mock(return_value='0'*32)
        c = self._makeit(f, ciph, mode='wb')
        d = '0' * 100
        with c:
            for _ in range(5):
                c.write(d)
        self.assertEqual(532, len(f.getvalue()))

    def test_read_context_exc(self):
        from lacore.exceptions import DecryptionError
        import StringIO
        f = StringIO.StringIO()
        ciph = self._makecipher()
        c = self._makeit(f, ciph)
        with ExpectedException(DecryptionError):
            with c:
                raise Exception()

    def test_hashobj(self):
        import StringIO
        from hashlib import md5
        f = StringIO.StringIO()
        d = '0' * 100
        ciph = self._makecipher()
        ciph.flush = Mock(return_value='0'*32)
        c = self._makeit(f, ciph, mode='wb', hashobj=md5())
        with c:
            for _ in range(4):
                c.write(d)
        self.assertEqual("93e84d29d7eb3ed55bd9adc243b06904",
                         c.hashobj.digest().encode('hex'))



from __future__ import unicode_literals
import platform

try:
    from lacore.version import __version__
except ImportError:
    __version__ = "0.1.3.1"


def get_client_info(terse=False, withplatform=True):
    if terse is True:
        el = ["lacore"]
    else:
        el = ["Longaccess Python SDK"]
    el.append(__version__)
    if withplatform is True:
        el.append(platform.platform(True, terse))
        el.append("py"+platform.python_version())
        el.append(platform.machine())
    return " ".join(el)
# vim: et:sw=4:ts=4

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from tempfile import NamedTemporaryFile
from zipfile import ZipFile, ZIP_DEFLATED
from shutil import copyfileobj
from contextlib import contextmanager
from . import Archiver

try:
    import zipstream
except ImportError:
    zipstream = None


class ZipArchiver(Archiver):
    def __init__(self, name=None, tmpdir=None, **kwargs):
        super(ZipArchiver, self).__init__(**kwargs)
        if name is None:
            name = ''
        self.name = name
        self.tmpdir = tmpdir

    @contextmanager
    def _temp_file(self, fdst):
        with NamedTemporaryFile(prefix=self.name, dir=self.tmpdir) as zf:
            yield zf
            zf.flush()
            zf.seek(0)
            with fdst as dst:
                copyfileobj(zf, dst, 1024)

    def args(self, items, cb=None):
        for item in items:
            if cb is not None:
                cb(item, item)
            yield ((item,), {})

    def archive(self, items, dst, cb=None):

        if zipstream is None:
            # do it in two passes now as vanilla zipfile
            # can't easily handle streaming
            dst = self._temp_file(dst)

        with dst as zf:
            zipargs = {'mode': 'w', 'compression': ZIP_DEFLATED,
                       'allowZip64': True}
            if zipstream is None:
                with ZipFile(zf, **zipargs) as zpf:
                    for args, kwargs in self.args(items, cb):
                        yield zpf.write(*args, **kwargs)
            else:
                with zipstream.ZipFile(**zipargs) as zpf:
                    zpf.paths_to_write = self.args(items, cb)
                    for chunk in zpf:
                        yield zf.write(chunk)



from testtools import TestCase


class ExceptionsTest(TestCase):
    def setUp(self):
        super(ExceptionsTest, self).setUp()

    def tearDown(self):
        super(ExceptionsTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacore.exceptions import BaseAppException
        return BaseAppException(*args, **kw)

    def test_exception(self):
        assert self._makeit(Exception())

    def test_str(self):
        exc = self._makeit(Exception("lala"))
        exc.msg = "foobar"
        self.assertEqual("foobar", str(exc))



from lacore.log import getLogger
from lacore.adf.elements import Links
from lacore.archive.urls import UrlArchiver
from lacore.archive.stream import StreamArchiver
from lacore.storage.s3 import S3Connection
from . import Dumper
from StringIO import StringIO


class BaseS3Dumper(Dumper):

    def __init__(self, key=None, retries=4, connection=None, **kwargs):
        if 'chunk' not in kwargs:
            kwargs['chunk'] = 10 * 1024 ** 2
        elif kwargs['chunk'] < 5242880:
            raise ValueError("S3Dumper chunk must be at least 5242880")
        super(BaseS3Dumper, self).__init__(**kwargs)
        self.connection = connection
        self.key = key
        if self.connection is None:
            self.connection = S3Connection()
        self.etags = []
        self.retries = retries
        self.seq = 0
        self.upload = None

    def update(self, result):
        super(BaseS3Dumper, self).update(result)
        if 'links' not in self.docs:
            self.docs['links'] = Links()
        if self.upload is not None:
            key = self.connection.complete_multipart(self.upload, self.etags)
            self.docs['links'].upload = (key.key_name, key.etag)

    @property
    def bucket(self):
        return self.connection.getbucket()

    def write(self, data):
        sz = len(data)
        fp = StringIO(data)
        for attempt in range(self.retries):
            fp.seek(0)
            getLogger().debug("attempt %d/%d to transfer part %d",
                              attempt, self.retries, self.seq)

            try:
                key = self.upload.upload_part_from_file(
                    fp=fp,
                    part_num=self.seq+1,
                    size=sz
                )
                getLogger().debug("succesfully uploaded part %d: %s",
                                  self.seq, key)
                self.seq += 1
                self.etags.append(key.etag)
                break
            except Exception as exc:
                getLogger().debug("exception while uploading part %d",
                                  self.seq, exc_info=True)

                if attempt == self.retries-1:
                    raise exc

    def __enter__(self):
        if self.upload is None:
            self.upload = self.connection.newupload(self.key)
        return self

    def __exit__(self, eType, eValue, eTrace):
        super(BaseS3Dumper, self).__exit__(eType, eValue, eTrace)
        if eType is not None:
            getLogger().debug("cancelling upload..", exc_info=True)
            if self.upload is not None:
                self.upload.cancel_upload()


class S3Dumper(BaseS3Dumper, UrlArchiver):
    pass


class S3StreamDumper(BaseS3Dumper, StreamArchiver):
    pass



import dateutil.parser
import dateutil.tz
from dateutil.relativedelta import relativedelta as date_delta
from datetime import datetime


def parse_timestamp(d):
    if not d:
        return d
    try:
        if not isinstance(d, datetime):
            d = dateutil.parser.parse(d)
        if d.utcoffset() is not None:
            return d.astimezone(dateutil.tz.tzutc())
        return d.replace(tzinfo=dateutil.tz.tzutc())
    except TypeError:
        raise ValueError("invalid time stamp: " + str(d))


def remaining_time(d):
    if d.utcoffset() is not None:
        return d - datetime.now(dateutil.tz.tzutc())
    return d - datetime.utcnow()


def today():
    return datetime.now(dateutil.tz.tzutc()).replace(microsecond=0)


def epoch():
    """ return a time-zone aware timestamp for the Unix epoch """
    return datetime.utcfromtimestamp(0).replace(tzinfo=dateutil.tz.tzutc())


def later(d, **args):
    return d + date_delta(**args)


def format_timestamp(d, tz=None):
    utc = dateutil.tz.tzutc()

    if not d.utcoffset():
        d = d.replace(tzinfo=utc)
    if tz and not tz.utcoffset():
        tz = None
    if tz:
        d = d.astimezone(tz)
    else:
        d = d.astimezone(utc)
    return d.replace(tzinfo=tz).isoformat()



from lacore.cipher import cipher_modes, new_key
from lacore.date import today, later
from .serialize import BaseYAMLObject, add_path_resolver


class Archive(BaseYAMLObject):
    """
    >>> import sys
    >>> meta = Meta('zip', 'aes-256-ctr', size=1024, created='now')
    >>> archive = Archive('title', meta, tags=['foo', 'bar'])
    >>> archive.title
    'title'
    >>> archive.description
    >>> archive.meta.cipher
    'aes-256-ctr'
    >>> archive.meta.size
    1024
    >>> archive.meta.created
    'now'
    >>> archive.tags[1]
    'bar'
    >>> archive.description = 'what a wonderful dataset'
    >>> archive.description
    'what a wonderful dataset'
    >>> Archive('title', meta, description='worthless junk').description
    'worthless junk'
    >>> import pyaml
    >>> pyaml.dump(Archive('foo', meta), sys.stdout)
    !archive
    meta: !meta
      cipher: aes-256-ctr
      created: now
      format: zip
      size: 1024
    title: foo
    """
    yaml_tag = u'!archive'
    title = None
    description = None
    tags = None
    meta = None

    def __init__(self, title, meta, description=None, tags=[]):
        self.title = title
        if description:
            self.description = description
        if len(tags):
            self.tags = tags
        if not hasattr(meta, 'format'):
            raise ValueError("invalid meta: " + str(meta))
        self.meta = meta


class Auth(BaseYAMLObject):
    """
    >>> from lacore.adf.persist import load_archive
    >>> with open('t/data/certs/2013-10-13-foobar.adf') as f:
    ...     doc=load_archive(f)['auth']
    ...     auth = Auth(sha512=doc.sha512)
    ...     auth.sha512.encode('hex')[:10]
    'd5ad8f4cb5'
    """
    yaml_tag = u'!auth'

    def __init__(self, *args, **kwargs):
        self.__dict__.update(kwargs)


class Meta(BaseYAMLObject):
    """
    >>> meta = Meta('zip', 'aes-256-ctr', size=1024)
    >>> meta.format
    'zip'
    >>> meta.cipher
    'aes-256-ctr'
    >>> meta.size
    1024
    >>> meta = Meta('zip', 'aes-256-ctr')
    >>> meta.size
    >>> meta.size = 1024
    >>> meta.size
    1024
    """
    yaml_tag = u'!meta'
    size = None
    format = None
    cipher = None
    email = None
    name = None
    created = None

    def __init__(self, format, cipher, size=None, created=None,
                 email=None, name=None):
        self.format = format
        self.cipher = cipher
        if size:
            self.size = size
        if email:
            self.email = email
        if name:
            self.name = name
        if created:
            self.created = created
        else:
            self.created = today()


add_path_resolver(u'!meta', ["meta"])


class Format(BaseYAMLObject):
    yaml_tag = u'!format'


add_path_resolver(u'!format', ["meta", "format"])


class Links(BaseYAMLObject):
    """
    >>> links = Links(local='http://foo.bar.com')
    >>> links.download
    >>> links.upload
    >>> links.local
    'http://foo.bar.com'
    >>> from StringIO import StringIO
    >>> out = StringIO()
    >>> from lacore.adf.persist import make_adf
    >>> make_adf(links, out=out, pretty=True)
    >>> print out.getvalue()
    !links
    local: http://foo.bar.com
    <BLANKLINE>
    """
    yaml_tag = u'!links'
    download = None
    local = None
    upload = None
    """
    This attribute is undocumented in the ADF format, because it only makes
    sense in the context of a specific client. It contains the URL of the
    pending upload API resource.
    """

    def __init__(self, download=None, local=None, upload=None):
        if download:
            # deprecated, but for compatibility with early client
            self.download = download
        if local:
            self.local = local
        if upload:
            self.upload = upload


class Cipher(BaseYAMLObject):
    """
    >>> cipher = Cipher('aes-256-ctr', key=2, input='\1\2\3\4'*4)
    >>> cipher.mode
    'aes-256-ctr'
    >>> cipher.key
    2
    >>> cipher.input.encode('hex')
    '01020304010203040102030401020304'
    >>> Cipher('aes-256-ctr', key='foo')
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "lacore/adf.py", line 70, in __init__
    ValueError: key must be integer
    >>> Cipher('wtf')
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "lacore/adf.py", line 68, in __init__
    ValueError: wtf not recognized
    >>> cipher = Cipher('aes-256-ctr')
    >>> hasattr(cipher, 'key')
    False
    >>> hasattr(cipher, 'input')
    False
    >>> cipher = Cipher('aes-256-ctr', key=1)
    >>> hasattr(cipher, 'key')
    False
    """

    yaml_tag = u'!cipher'

    def __init__(self, mode, key=1, input=None):
        if mode not in cipher_modes():
            raise ValueError("{} not recognized".format(mode))
        if not isinstance(key, int):
            raise ValueError("key must be integer")
        self.mode = mode
        if key > 1:
            self.key = key
        if input:
            self.input = input

    def __setstate__(self, d):
        assert d['mode'] in cipher_modes()
        self.__dict__.update(d)


add_path_resolver(u'!cipher', ["meta", "cipher"])


class DerivedKey(BaseYAMLObject):
    yaml_tag = u'!key'


add_path_resolver(u'!key', ["keys", None])


class Certificate(BaseYAMLObject):
    """
    >>> import sys
    >>> cert = Certificate(chr(255)*16)
    >>> import pyaml
    >>> pyaml.dump(cert, sys.stdout)
    !certificate
    key: !!binary |
      /////////////////////w==
    """
    yaml_tag = u'!certificate'

    def __init__(self, key=None):
        if not key:
            self.key = new_key(256)
        else:
            self.key = key


class MAC(BaseYAMLObject):
    yaml_tag = u'!mac'


add_path_resolver(u'!mac', ["mac"])


class Signature(BaseYAMLObject):
    """
    >>> import sys
    >>> from datetime import datetime
    >>> sig = Signature('1', 'http://baz.com', datetime.utcfromtimestamp(0))
    >>> import pyaml
    >>> pyaml.dump(sig, sys.stdout)
    !signature
    aid: '1'
    created: 1970-01-01 00:00:00
    expires: 2000-01-01 00:00:00
    uri: http://baz.com
    """
    yaml_tag = u'!signature'
    aid = None
    uri = None
    created = None
    expires = None
    creator = None

    def __init__(self, aid, uri, created=None, expires=None, creator=None):
        self.aid = aid
        self.uri = uri
        if created:
            self.created = created
        else:
            self.created = today()

        if expires:
            self.expires = expires
        else:
            self.expires = later(self.created, years=30)

        if creator:
            self.creator = creator


add_path_resolver(u'!signature', ["signature"])
